# Copyright (c) 2011-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Dict, Iterator, List, cast, Optional
import logging
from ezdxf.lldxf.const import DXFKeyError, DXFValueError, DXFInternalEzdxfError
from ezdxf.lldxf.const import (
    MODEL_SPACE_R2000,
    PAPER_SPACE_R2000,
    TMP_PAPER_SPACE_NAME,
)
from ezdxf.lldxf.validator import is_valid_table_name
from .layout import Layout, Modelspace, Paperspace
from ezdxf.entities import DXFEntity

if TYPE_CHECKING:
    from ezdxf.eztypes import Dictionary, Drawing, Auditor, DXFLayout

logger = logging.getLogger("ezdxf")


def key(name: str) -> str:
    """AutoCAD uses case insensitive layout names, but stores the name case
    sensitive."""
    return name.upper()


MODEL = key("Model")


class Layouts:
    def __init__(self, doc: "Drawing"):
        """Default constructor. (internal API)"""
        self.doc = doc
        # Store layout names in normalized form: key(name)
        self._layouts: Dict[str, Layout] = {}
        # key: layout name as original case sensitive string; value: DXFLayout()
        self._dxf_layouts: "Dictionary" = cast(
            "Dictionary", self.doc.rootdict["ACAD_LAYOUT"]
        )

    @classmethod
    def setup(cls, doc: "Drawing"):
        """Constructor from scratch. (internal API)"""
        layouts = Layouts(doc)
        layouts.setup_modelspace()
        layouts.setup_paperspace()
        return layouts

    def __len__(self) -> int:
        """Returns count of existing layouts, including the modelspace
        layout."""
        return len(self._layouts)

    def __contains__(self, name: str) -> bool:
        """Returns ``True`` if layout `name` exist."""
        assert isinstance(name, str), type(str)
        return key(name) in self._layouts

    def __iter__(self) -> Iterator["Layout"]:
        """Returns iterable of all layouts as :class:`~ezdxf.layouts.Layout`
        objects, including the modelspace layout.
        """
        return iter(self._layouts.values())

    def _add_layout(self, name: str, layout: Layout):
        layout.dxf.name = name
        self._layouts[key(name)] = layout
        self._dxf_layouts[name] = layout.dxf_layout

    def _discard(self, layout: "Layout"):
        name = layout.name
        self._dxf_layouts.discard(name)
        del self._layouts[key(name)]

    def setup_modelspace(self):
        """Modelspace setup. (internal API)"""
        self._new_special(
            Modelspace, "Model", MODEL_SPACE_R2000, dxfattribs={"taborder": 0}
        )

    def setup_paperspace(self):
        """First layout setup. (internal API)"""
        self._new_special(
            Paperspace, "Layout1", PAPER_SPACE_R2000, dxfattribs={"taborder": 1}
        )

    def _new_special(
        self, cls, name: str, block_name: str, dxfattribs: dict
    ) -> "Layout":
        if name in self._layouts:
            raise DXFValueError(f'Layout "{name}" already exists')
        dxfattribs["owner"] = self._dxf_layouts.dxf.handle
        layout = cls.new(name, block_name, self.doc, dxfattribs=dxfattribs)
        self._add_layout(name, layout)
        return layout

    def unique_paperspace_name(self) -> str:
        """Returns a unique paperspace name. (internal API)"""
        blocks = self.doc.blocks
        count = 0
        while "*Paper_Space%d" % count in blocks:
            count += 1
        return "*Paper_Space%d" % count

    def new(self, name: str, dxfattribs: dict = None) -> Paperspace:
        """Returns a new :class:`~ezdxf.layouts.Paperspace` layout.

        Args:
            name: layout name as shown in tabs in :term:`CAD` applications
            dxfattribs: additional DXF attributes for the
                :class:`~ezdxf.entities.layout.DXFLayout` entity

        Raises:
            DXFValueError: Invalid characters in layout name.
            DXFValueError: Layout `name` already exist.

        """
        assert isinstance(name, str), type(str)
        if not is_valid_table_name(name):
            raise DXFValueError("Layout name contains invalid characters.")

        if name in self:
            raise DXFValueError(f'Layout "{name}" already exist.')

        dxfattribs = dict(dxfattribs or {})  # copy attribs
        dxfattribs["owner"] = self._dxf_layouts.dxf.handle
        dxfattribs.setdefault("taborder", len(self._layouts) + 1)
        block_name = self.unique_paperspace_name()
        layout = Paperspace.new(
            name, block_name, self.doc, dxfattribs=dxfattribs
        )
        # Default extents are ok!
        # Reset limits to (0, 0) and (paper width, paper height)
        layout.reset_limits()
        self._add_layout(name, layout)
        return layout  # type: ignore

    @classmethod
    def load(cls, doc: "Drawing") -> "Layouts":
        """Constructor if loading from file. (internal API)"""
        layouts = cls(doc)
        layouts.setup_from_rootdict()

        # DXF R12: block/block_record for *Model_Space and *Paper_Space
        # already exist:
        if len(layouts) < 2:  # restore missing DXF Layouts
            layouts.restore("Model", MODEL_SPACE_R2000, taborder=0)
            layouts.restore("Layout1", PAPER_SPACE_R2000, taborder=1)
        return layouts

    def restore(self, name: str, block_record_name: str, taborder: int) -> None:
        """Restore layout from block if DXFLayout does not exist.
        (internal API)"""
        if name in self:
            return
        block_layout = self.doc.blocks.get(block_record_name)
        self._new_from_block_layout(name, block_layout, taborder)

    def _new_from_block_layout(
        self, name, block_layout, taborder: int
    ) -> "Layout":
        dxfattribs = {
            "owner": self._dxf_layouts.dxf.handle,
            "name": name,
            "block_record_handle": block_layout.block_record_handle,
            "taborder": taborder,
        }
        dxf_layout = cast(
            "DXFLayout",
            self.doc.objects.new_entity("LAYOUT", dxfattribs=dxfattribs),
        )
        if key(name) == MODEL:
            layout = Modelspace.load(dxf_layout, self.doc)
        else:
            layout = Paperspace.load(dxf_layout, self.doc)
        self._add_layout(name, layout)
        return layout

    def setup_from_rootdict(self) -> None:
        """Setup layout manger from root dictionary. (internal API)"""
        layout: Layout
        for name, dxf_layout in self._dxf_layouts.items():
            if isinstance(dxf_layout, str):
                logger.debug(
                    f"ignore missing LAYOUT(#{dxf_layout}) entity '{name}'"
                )
                continue
            if key(name) == MODEL:
                layout = Modelspace(dxf_layout, self.doc)
            else:
                layout = Paperspace(dxf_layout, self.doc)
            # assert name == layout.dxf.name
            self._layouts[key(name)] = layout

    def modelspace(self) -> Modelspace:
        """Returns the :class:`~ezdxf.layouts.Modelspace` layout."""
        return cast(Modelspace, self.get("Model"))

    def names(self) -> List[str]:
        """Returns a list of all layout names, all names in original case
        sensitive form."""
        return [layout.name for layout in self._layouts.values()]

    def get(self, name: Optional[str]) -> "Layout":
        """Returns :class:`~ezdxf.layouts.Layout` by `name`, case insensitive
        "Model" == "MODEL".

        Args:
            name: layout name as shown in tab, e.g. ``'Model'`` for modelspace

        """
        name = name or self.names_in_taborder()[1]  # first paperspace layout
        return self._layouts[key(name)]

    def rename(self, old_name: str, new_name: str) -> None:
        """Rename a layout from `old_name` to `new_name`.
        Can not rename layout ``'Model'`` and the new name of a layout must
        not exist.

        Args:
            old_name: actual layout name, case insensitive
            new_name: new layout name, case insensitive

        Raises:
            DXFValueError: try to rename ``'Model'``
            DXFValueError: Layout `new_name` already exist.

        """
        assert isinstance(old_name, str), type(old_name)
        assert isinstance(new_name, str), type(new_name)
        if key(old_name) == MODEL:
            raise DXFValueError("Can not rename model space.")
        if new_name in self:
            raise DXFValueError(f'Layout "{new_name}" already exist.')
        if old_name not in self:
            raise DXFValueError(f'Layout "{old_name}" does not exist.')

        layout = self.get(old_name)
        self._discard(layout)
        layout.rename(new_name)
        self._add_layout(new_name, layout)

    def names_in_taborder(self) -> List[str]:
        """Returns all layout names in tab order as shown in :term:`CAD`
        applications."""
        names = [
            (layout.dxf.taborder, layout.name)
            for layout in self._layouts.values()
        ]
        return [name for order, name in sorted(names)]

    def get_layout_for_entity(self, entity: "DXFEntity") -> "Layout":
        """Returns the owner layout for a DXF `entity`."""
        owner = entity.dxf.owner
        if owner is None:
            raise DXFKeyError("No associated layout, owner is None.")
        return self.get_layout_by_key(entity.dxf.owner)

    def get_layout_by_key(self, layout_key: str) -> "Layout":
        """Returns a layout by its `layout_key`. (internal API)"""
        assert isinstance(layout_key, str), type(layout_key)
        try:
            block_record = self.doc.entitydb[layout_key]
            dxf_layout = self.doc.entitydb[block_record.dxf.layout]
        except KeyError:
            raise DXFKeyError(f'Layout with key "{layout_key}" does not exist.')
        return self.get(dxf_layout.dxf.name)

    def get_active_layout_key(self):
        """Returns layout kay for the active paperspace layout.
        (internal API)"""
        active_layout_block_record = self.doc.block_records.get(
            PAPER_SPACE_R2000
        )
        return active_layout_block_record.dxf.handle

    def set_active_layout(self, name: str) -> None:
        """Set layout `name` as active paperspace layout."""
        assert isinstance(name, str), type(name)
        if key(name) == MODEL:  # reserved layout name
            raise DXFValueError("Can not set model space as active layout")
        # raises KeyError if layout 'name' does not exist
        new_active_layout = self.get(name)
        old_active_layout_key = self.get_active_layout_key()
        if old_active_layout_key == new_active_layout.layout_key:
            return  # layout 'name' is already the active layout

        blocks = self.doc.blocks
        new_active_paper_space_name = new_active_layout.block_record_name

        blocks.rename_block(PAPER_SPACE_R2000, TMP_PAPER_SPACE_NAME)
        blocks.rename_block(new_active_paper_space_name, PAPER_SPACE_R2000)
        blocks.rename_block(TMP_PAPER_SPACE_NAME, new_active_paper_space_name)

    def delete(self, name: str) -> None:
        """Delete layout `name` and destroy all entities in that layout.

        Args:
            name (str): layout name as shown in tabs

        Raises:
            DXFKeyError: if layout `name` do not exists
            DXFValueError: deleting modelspace layout is not possible
            DXFValueError: deleting last paperspace layout is not possible

        """
        assert isinstance(name, str), type(name)
        if key(name) == MODEL:
            raise DXFValueError("Can not delete modelspace layout.")

        layout = self.get(name)
        if len(self) < 3:
            raise DXFValueError("Can not delete last paperspace layout.")
        if layout.layout_key == self.get_active_layout_key():
            # Layout `name` is the active layout:
            for layout_name in self._layouts:
                # Set any other paperspace layout as active layout
                if layout_name not in (key(name), MODEL):
                    self.set_active_layout(layout_name)
                    break
        self._discard(layout)
        layout.destroy()

    def active_layout(self) -> Paperspace:
        """Returns the active paperspace layout."""
        for layout in self:
            if layout.is_active_paperspace:
                return cast(Paperspace, layout)
        raise DXFInternalEzdxfError("No active paperspace layout found.")

    def audit(self, auditor: "Auditor"):
        from ezdxf.audit import AuditError

        doc = auditor.doc

        # Find/remove orphaned LAYOUT objects:
        layouts = (o for o in doc.objects if o.dxftype() == "LAYOUT")
        for layout in layouts:
            name = layout.dxf.get("name")
            if name not in self:
                auditor.fixed_error(
                    code=AuditError.ORPHANED_LAYOUT_ENTITY,
                    message=f'Removed orphaned {str(layout)} "{name}"',
                )
                doc.objects.delete_entity(layout)

        # Find/remove orphaned paperspace BLOCK_RECORDS named: *Paper_Space...
        psp_br_handles = {
            br.dxf.handle
            for br in doc.block_records
            if br.dxf.name.lower().startswith("*paper_space")
        }
        psp_layout_br_handles = {
            layout.dxf.block_record_handle
            for layout in self._layouts.values()
            if key(layout.name) != MODEL
        }
        mismatch = psp_br_handles.difference(psp_layout_br_handles)
        if len(mismatch):
            for handle in mismatch:
                br = doc.entitydb.get(handle)
                name = br.dxf.get("name")
                auditor.fixed_error(
                    code=AuditError.ORPHANED_PAPER_SPACE_BLOCK_RECORD_ENTITY,
                    message=f'Removed orphaned layout {str(br)} "{name}"',
                )
                if name in doc.blocks:
                    doc.blocks.delete_block(name)
                else:
                    doc.block_records.remove(name)
        # Does not check the LAYOUT content this is done in the BlockSection,
        # because the content of layouts is stored as blocks.



# Copyright (c) 2020-2021, Manfred Moitzi
# License: MIT License
from typing import (
    TYPE_CHECKING,
    Optional,
    Iterable,
    Tuple,
    List,
    Set,
    Dict,
    cast,
    Sequence,
    Any,
)
import sys
import struct
import math
from enum import IntEnum
from itertools import repeat
from ezdxf.lldxf import const
from ezdxf.tools.binarydata import bytes_to_hexstr, ByteStream, BitStream
from ezdxf import colors
from ezdxf.math import (
    Vec3,
    Matrix44,
    Z_AXIS,
    ConstructionCircle,
    ConstructionArc,
    OCS,
    UCS,
    X_AXIS,
)
from ezdxf.entities import factory
import logging

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        Tags,
        TagWriter,
        Drawing,
        DXFGraphic,
        Polymesh,
        Polyface,
        Polyline,
        Hatch,
        LWPolyline,
    )

logger = logging.getLogger("ezdxf")

CHUNK_SIZE = 127


def load_proxy_graphic(
    tags: "Tags", length_code: int = 160, data_code: int = 310
) -> Optional[bytes]:
    binary_data = [
        tag.value
        for tag in tags.pop_tags(codes=(length_code, data_code))
        if tag.code == data_code
    ]
    return b"".join(binary_data) if len(binary_data) else None  # type: ignore


def export_proxy_graphic(
    data: bytes,
    tagwriter: "TagWriter",
    length_code: int = 160,
    data_code: int = 310,
) -> None:
    # Do not export proxy graphic for DXF R12 files
    assert tagwriter.dxfversion > const.DXF12

    length = len(data)
    if length == 0:
        return

    tagwriter.write_tag2(length_code, length)
    index = 0
    while index < length:
        hex_str = bytes_to_hexstr(data[index : index + CHUNK_SIZE])
        tagwriter.write_tag2(data_code, hex_str)
        index += CHUNK_SIZE


def has_prim_traits(flags: int) -> bool:
    return bool(flags & 0xFFFF)


def prims_have_colors(flags: int) -> bool:
    return bool(flags & 0x0001)


def prims_have_layers(flags: int) -> bool:
    return bool(flags & 0x0002)


def prims_have_linetypes(flags: int) -> bool:
    return bool(flags & 0x0004)


def prims_have_markers(flags: int) -> bool:
    return bool(flags & 0x0020)


def prims_have_visibilities(flags: int) -> bool:
    return bool(flags & 0x0040)


def prims_have_normals(flags: int) -> bool:
    return bool(flags & 0x0080)


def prims_have_orientation(flags: int) -> bool:
    return bool(flags & 0x0400)


TRAIT_TESTER = {
    "colors": (prims_have_colors, "RL"),
    "layers": (prims_have_layers, "RL"),
    "linetypes": (prims_have_linetypes, "RL"),
    "markers": (prims_have_markers, "RL"),
    "visibilities": (prims_have_visibilities, "RL"),
    "normals": (prims_have_normals, "3RD"),
}


def read_prim_traits(
    bs: ByteStream, types: Sequence[str], prim_flags: int, count: int
) -> Dict:
    def read_float_list():
        return [bs.read_long() for _ in range(count)]

    def read_vertices():
        return [Vec3(bs.read_vertex()) for _ in range(count)]

    data = dict()
    for t in types:
        test_trait, data_type = TRAIT_TESTER[t]
        if test_trait(prim_flags):
            if data_type == "3RD":
                data[t] = read_vertices()
            elif data_type == "RL":
                data[t] = read_float_list()
            else:
                raise TypeError(data_type)
    return data


def read_mesh_traits(
    bs: ByteStream, edge_count: int, face_count: int, vertex_count: int
):
    # Traits data format:
    # all entries are optional
    # traits: Dict[str, Dict]
    #     "edges": Dict[str, List]
    #         "colors": List[int]
    #         "layers": List[int] as layer ids
    #         "linetypes": List[int] as linetype ids
    #         "markers": List[int]
    #         "visibilities": List[int]
    #     "faces": Dict[str, List]
    #         "colors": List[int]
    #         "layers": List[int] as layer ids
    #         "markers": List[int]
    #         "normals": List[Vec3]
    #         "visibilities": List[int]
    #     "vertices": Dict
    #         "normals": List[Vec3]
    #         "orientation": bool
    traits = dict()
    edge_flags = bs.read_long()
    if has_prim_traits(edge_flags):
        traits["edges"] = read_prim_traits(
            bs,
            ["colors", "layers", "linetypes", "markers", "visibilities"],
            edge_flags,
            edge_count,
        )
    face_flags = bs.read_long()
    if has_prim_traits(face_flags):
        traits["faces"] = read_prim_traits(
            bs,
            ["colors", "layers", "markers", "normals", "visibilities"],
            face_flags,
            face_count,
        )

    # Note: DXF entities PolyFaceMesh and Mesh do not support vertex normals!
    # disable useless reading process by vertex_count = 0
    if vertex_count > 0:
        vertex_flags = bs.read_long()
        if has_prim_traits(vertex_flags):
            vertices = dict()
            if prims_have_normals(vertex_flags):
                vertices["normals"] = [
                    Vec3(bs.read_vertex()) for _ in range(vertex_count)
                ]
            if prims_have_orientation(vertex_flags):
                vertices["orientation"] = bool(bs.read_long())  # type: ignore
            traits["vertices"] = vertices
    return traits


class ProxyGraphicTypes(IntEnum):
    EXTENTS = 1
    CIRCLE = 2
    CIRCLE_3P = 3
    CIRCULAR_ARC = 4
    CIRCULAR_ARC_3P = 5
    POLYLINE = 6
    POLYGON = 7
    MESH = 8
    SHELL = 9
    TEXT = 10
    TEXT2 = 11
    XLINE = 12
    RAY = 13
    ATTRIBUTE_COLOR = 14
    UNUSED_15 = 15
    ATTRIBUTE_LAYER = 16
    UNUSED_17 = 17
    ATTRIBUTE_LINETYPE = 18
    ATTRIBUTE_MARKER = 19
    ATTRIBUTE_FILL = 20
    UNUSED_21 = 21
    ATTRIBUTE_TRUE_COLOR = 22
    ATTRIBUTE_LINEWEIGHT = 23
    ATTRIBUTE_LTSCALE = 24
    ATTRIBUTE_THICKNESS = 25
    ATTRIBUTE_PLOT_STYLE_NAME = 26
    PUSH_CLIP = 27
    POP_CLIP = 28
    PUSH_MATRIX = 29
    PUSH_MATRIX2 = 30
    POP_MATRIX = 31
    POLYLINE_WITH_NORMALS = 32
    LWPOLYLINE = 33
    ATTRIBUTE_MATERIAL = 34
    ATTRIBUTE_MAPPER = 35
    UNICODE_TEXT = 36
    UNKNOWN_37 = 37
    UNICODE_TEXT2 = 38


class ProxyGraphic:
    def __init__(self, data: bytes, doc: "Drawing" = None):
        self._doc = doc
        self._factory = factory.new
        self._buffer: bytes = data
        self._index: int = 8
        self.dxfversion = doc.dxfversion if doc else "AC1015"
        self.color: int = const.BYLAYER
        self.layer: str = "0"
        self.linetype: str = "BYLAYER"
        self.marker_index: int = 0
        self.fill: bool = False
        self.true_color: Optional[int] = None
        self.lineweight: int = const.LINEWEIGHT_DEFAULT
        self.ltscale: float = 1.0
        self.thickness: float = 0.0
        # Layer list in storage order
        self.layers: List[str] = []
        # Linetypes list in storage order
        self.linetypes: List[str] = []
        # List of text styles, with font name as key
        self.textstyles: Dict[str, str] = dict()
        self.required_fonts: Set[str] = set()
        self.matrices: List[Matrix44] = []

        if self._doc:
            self.layers = list(layer.dxf.name for layer in self._doc.layers)
            self.linetypes = list(
                linetype.dxf.name for linetype in self._doc.linetypes
            )
            self.textstyles = {
                style.dxf.font: style.dxf.name for style in self._doc.styles
            }

    def info(self) -> Iterable[Tuple[int, int, str]]:
        index = self._index
        buffer = self._buffer
        while index < len(buffer):
            size, type_ = struct.unpack_from("<2L", self._buffer, offset=index)
            try:
                name = ProxyGraphicTypes(type_).name
            except ValueError:
                name = f"UNKNOWN_TYPE_{type_}"
            yield index, size, name
            index += size

    def __virtual_entities__(self) -> Iterable["DXFGraphic"]:
        """Implements the SupportsVirtualEntities protocol."""

        def transform(entity):
            if self.matrices:
                return entity.transform(self.matrices[-1])
            else:
                return entity

        index = self._index
        buffer = self._buffer
        while index < len(buffer):
            size, type_ = struct.unpack_from("<2L", self._buffer, offset=index)
            try:
                name = ProxyGraphicTypes(type_).name.lower()
            except ValueError:
                logger.debug(f"Unsupported Type Code: {type_}")
                index += size
                continue
            method = getattr(self, name, None)
            if method:
                result = method(self._buffer[index + 8 : index + size])
                if isinstance(result, tuple):
                    for entity in result:
                        yield transform(entity)
                elif result:
                    yield transform(result)
                if result:  # reset fill after each graphic entity
                    self.fill = False
            else:
                logger.debug(f"Unsupported feature ProxyGraphic.{name}()")
            index += size

    def virtual_entities(self) -> Iterable["DXFGraphic"]:
        return self.__virtual_entities__()

    def push_matrix(self, data: bytes):
        values = struct.unpack("<16d", data)
        m = Matrix44(values)
        m.transpose()
        self.matrices.append(m)

    def pop_matrix(self, data: bytes):
        if self.matrices:
            self.matrices.pop()

    def reset_colors(self):
        self.color = const.BYLAYER
        self.true_color = None

    def attribute_color(self, data: bytes):
        self.reset_colors()
        self.color = struct.unpack("<L", data)[0]
        if self.color < 0 or self.color > 256:
            self.color = const.BYLAYER

    def attribute_layer(self, data: bytes):
        if self._doc:
            index = struct.unpack("<L", data)[0]
            if index < len(self.layers):
                self.layer = self.layers[index]

    def attribute_linetype(self, data: bytes):
        if self._doc:
            index = struct.unpack("<L", data)[0]
            if index < len(self.linetypes):
                self.linetype = self.linetypes[index]

    def attribute_marker(self, data: bytes):
        self.marker_index = struct.unpack("<L", data)[0]

    def attribute_fill(self, data: bytes):
        self.fill = bool(struct.unpack("<L", data)[0])

    def attribute_true_color(self, data: bytes):
        self.reset_colors()
        code, value = colors.decode_raw_color(struct.unpack("<L", data)[0])
        if code == colors.COLOR_TYPE_RGB:
            self.true_color = colors.rgb2int(value)  # type: ignore
        else:  # ACI colors, BYLAYER, BYBLOCK
            self.color = value  # type: ignore

    def attribute_lineweight(self, data: bytes):
        lw = struct.unpack("<L", data)[0]
        if lw > const.MAX_VALID_LINEWEIGHT:
            self.lineweight = max(lw - 0x100000000, const.LINEWEIGHT_DEFAULT)
        else:
            self.lineweight = lw

    def attribute_ltscale(self, data: bytes):
        self.ltscale = struct.unpack("<d", data)[0]

    def attribute_thickness(self, data: bytes):
        self.thickness = struct.unpack("<d", data)[0]

    def circle(self, data: bytes):
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        attribs["center"] = Vec3(bs.read_vertex())
        attribs["radius"] = bs.read_float()
        attribs["extrusion"] = bs.read_vertex()
        return self._factory("CIRCLE", dxfattribs=attribs)

    def circle_3p(self, data: bytes):
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        p1 = Vec3(bs.read_vertex())
        p2 = Vec3(bs.read_vertex())
        p3 = Vec3(bs.read_vertex())
        circle = ConstructionCircle.from_3p(p1, p2, p3)
        attribs["center"] = circle.center
        attribs["radius"] = circle.radius
        return self._factory("CIRCLE", dxfattribs=attribs)

    def circular_arc(self, data: bytes):
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        center = Vec3(bs.read_vertex())  # in WCS
        attribs["radius"] = bs.read_float()
        normal = Vec3(bs.read_vertex())  # UCS z-axis
        start_vec = Vec3(bs.read_vertex())  # UCS x-axis
        # sweep angle around normal vector!
        sweep_angle = bs.read_float()  # in radians
        # arc_type = bs.read_long()  # unused yet - meaning?
        start_angle: float  # in degrees
        end_angle: float  # in degrees
        if not normal.isclose(Z_AXIS):
            # local UCS
            ucs = UCS(ux=start_vec, uz=normal)
            # target OCS
            ocs = OCS(normal)
            # convert start angle == UCS x-axis to OCS
            start_angle = ocs.from_wcs(ucs.to_wcs(X_AXIS)).angle_deg  # type: ignore
            # convert end angle to OCS
            end_vec = Vec3.from_angle(sweep_angle)
            end_angle = ocs.from_wcs(ucs.to_wcs(end_vec)).angle_deg  # type: ignore
            # setup OCS for ARC entity
            attribs["extrusion"] = normal
            # convert WCS center to OCS center
            center = ocs.from_wcs(center)
        else:
            start_angle = start_vec.angle_deg
            end_angle = start_angle + math.degrees(sweep_angle)
        attribs["center"] = center
        attribs["start_angle"] = start_angle
        attribs["end_angle"] = end_angle
        return self._factory("ARC", dxfattribs=attribs)

    def circular_arc_3p(self, data: bytes):
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        p1 = Vec3(bs.read_vertex())
        p2 = Vec3(bs.read_vertex())
        p3 = Vec3(bs.read_vertex())
        # arc_type = bs.read_long()  # unused yet
        arc = ConstructionArc.from_3p(p1, p3, p2)
        attribs["center"] = arc.center
        attribs["radius"] = arc.radius
        attribs["start_angle"] = arc.start_angle
        attribs["end_angle"] = arc.end_angle
        return self._factory("ARC", dxfattribs=attribs)

    def _filled_polygon(self, vertices, attribs):
        hatch = cast("Hatch", self._factory("HATCH", dxfattribs=attribs))
        hatch.paths.add_polyline_path(vertices, is_closed=True)
        return hatch

    def _polyline(self, vertices, *, close=False, normal=Z_AXIS):
        # Polyline without bulge values!
        # Current implementation ignores the normal vector!
        attribs = self._build_dxf_attribs()
        count = len(vertices)
        if count == 1 or (count == 2 and vertices[0].isclose(vertices[1])):
            attribs["location"] = vertices[0]
            return self._factory("POINT", dxfattribs=attribs)

        if self.fill and count > 2:
            polyline = self._filled_polygon(vertices, attribs)
        else:
            attribs["flags"] = const.POLYLINE_3D_POLYLINE
            polyline = cast(
                "Polyline", self._factory("POLYLINE", dxfattribs=attribs)
            )
            polyline.append_vertices(vertices)
            if close:
                polyline.close()
        return polyline

    def polyline_with_normals(self, data: bytes):
        # Polyline without bulge values!
        vertices, normal = self._load_vertices(data, load_normal=True)
        return self._polyline(vertices, normal=normal)

    def polyline(self, data: bytes):
        # Polyline without bulge values!
        vertices, normal = self._load_vertices(data, load_normal=False)
        return self._polyline(vertices)

    def polygon(self, data: bytes):
        # Polyline without bulge values!
        vertices, normal = self._load_vertices(data, load_normal=False)
        return self._polyline(vertices, close=True)

    def lwpolyline(self, data: bytes):
        # OpenDesign Specs LWPLINE: 20.4.85 Page 211
        logger.warning(
            "Untested proxy graphic entity: LWPOLYLINE - Need examples!"
        )
        bs = BitStream(data)
        flag: int = bs.read_bit_short()
        attribs = self._build_dxf_attribs()
        if flag & 4:
            attribs["const_width"] = bs.read_bit_double()
        if flag & 8:
            attribs["elevation"] = bs.read_bit_double()
        if flag & 2:
            attribs["thickness"] = bs.read_bit_double()
        if flag & 1:
            attribs["extrusion"] = Vec3(bs.read_bit_double(3))

        num_points = bs.read_bit_long()
        if flag & 16:
            num_bulges = bs.read_bit_long()
        else:
            num_bulges = 0

        if self.dxfversion >= "AC1024":  # R2010+
            vertex_id_count = bs.read_bit_long()
        else:
            vertex_id_count = 0

        if flag & 32:
            num_width = bs.read_bit_long()
        else:
            num_width = 0
        # ignore DXF R13/14 special vertex order

        vertices: List[Tuple[float, float]] = [bs.read_raw_double(2)]  # type: ignore
        prev_point = vertices[-1]
        for _ in range(num_points - 1):
            x = bs.read_bit_double_default(default=prev_point[0])  # type: ignore
            y = bs.read_bit_double_default(default=prev_point[1])  # type: ignore
            prev_point = (x, y)
            vertices.append(prev_point)
        bulges: List[float] = [bs.read_bit_double() for _ in range(num_bulges)]
        vertex_ids: List[int] = [
            bs.read_bit_long() for _ in range(vertex_id_count)
        ]
        widths: List[Tuple[float, float]] = [
            (bs.read_bit_double(), bs.read_bit_double())
            for _ in range(num_width)
        ]
        if len(bulges) == 0:
            bulges = list(repeat(0, num_points))
        if len(widths) == 0:
            widths = list(repeat((0, 0), num_points))
        points: List[Sequence[float]] = []
        for v, w, b in zip(vertices, widths, bulges):
            points.append((v[0], v[1], w[0], w[1], b))
        lwpolyline = cast(
            "LWPolyline", self._factory("LWPOLYLINE", dxfattribs=attribs)
        )
        lwpolyline.set_points(points)
        return lwpolyline

    def mesh(self, data: bytes):
        # Limitations of the PolyFacMesh entity:
        # - all VERTEX entities have to reside on the same layer
        # - does not support vertex normals
        # - all faces have the same color (no face record)

        logger.warning("Untested proxy graphic entity: MESH - Need examples!")
        bs = ByteStream(data)
        rows, columns = bs.read_struct("<2L")
        total_edge_count = (rows - 1) * columns + (columns - 1) * rows
        total_face_count = (rows - 1) * (columns - 1)
        total_vertex_count = rows * columns
        vertices = [Vec3(bs.read_vertex()) for _ in range(total_vertex_count)]

        traits = dict()
        try:
            traits = read_mesh_traits(
                bs, total_edge_count, total_face_count, vertex_count=0
            )
        except struct.error:
            logger.error(
                "Structure error while parsing traits for MESH proxy graphic"
            )
        if traits:
            # apply traits
            pass

        # create PolyMesh entity
        attribs = self._build_dxf_attribs()
        attribs["m_count"] = rows
        attribs["n_count"] = columns
        attribs["flags"] = const.POLYLINE_3D_POLYMESH
        polymesh = cast(
            "Polymesh", self._factory("POLYLINE", dxfattribs=attribs)
        )
        polymesh.append_vertices(vertices)
        return polymesh

    def shell(self, data: bytes):
        # Limitations of the PolyFacMesh entity:
        # - all VERTEX entities have to reside on the same layer
        # - does not support vertex normals
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        attribs["flags"] = const.POLYLINE_POLYFACE
        polyface = cast(
            "Polyface", self._factory("POLYLINE", dxfattribs=attribs)
        )
        total_vertex_count = bs.read_long()
        vertices = [Vec3(bs.read_vertex()) for _ in range(total_vertex_count)]
        face_entry_count = bs.read_long()
        faces = []
        read_count: int = 0
        total_face_count: int = 0
        total_edge_count: int = 0
        while read_count < face_entry_count:
            edge_count = abs(bs.read_signed_long())
            read_count += 1 + edge_count
            face_indices = [bs.read_long() for _ in range(edge_count)]
            face = [vertices[index] for index in face_indices]
            total_face_count += 1
            total_edge_count += edge_count
            faces.append(face)

        traits = dict()
        try:
            traits = read_mesh_traits(
                bs, total_edge_count, total_face_count, vertex_count=0
            )
        except struct.error:
            logger.error(
                "Structure error while parsing traits for SHELL proxy graphic"
            )
        polyface.append_faces(faces)
        if traits:
            face_traits = traits.get("faces")
            if face_traits:
                face_colors = face_traits.get("colors")
                if face_colors:
                    logger.warning(
                        "Untested proxy graphic feature for SHELL: "
                        "apply face colors - Need examples!"
                    )
                    assert isinstance(face_colors, list)
                    _apply_face_colors(polyface, face_colors)
        polyface.optimize()
        return polyface

    def text(self, data: bytes):
        return self._text(data, unicode=False)

    def unicode_text(self, data: bytes):
        return self._text(data, unicode=True)

    def _text(self, data: bytes, unicode: bool = False):
        bs = ByteStream(data)
        start_point = Vec3(bs.read_vertex())
        normal = Vec3(bs.read_vertex())
        text_direction = Vec3(bs.read_vertex())
        height, width_factor, oblique_angle = bs.read_struct("<3d")
        if unicode:
            text = bs.read_padded_unicode_string()
        else:
            text = bs.read_padded_string()
        attribs = self._build_dxf_attribs()
        attribs["insert"] = start_point
        attribs["text"] = text
        attribs["height"] = height
        attribs["width"] = width_factor
        attribs["rotation"] = text_direction.angle_deg
        attribs["oblique"] = math.degrees(oblique_angle)
        attribs["extrusion"] = normal
        return self._factory("TEXT", dxfattribs=attribs)

    def text2(self, data: bytes):
        bs = ByteStream(data)
        start_point = Vec3(bs.read_vertex())
        normal = Vec3(bs.read_vertex())
        text_direction = Vec3(bs.read_vertex())
        text = bs.read_padded_string()
        ignore_length_of_string, raw = bs.read_struct("<2l")
        (
            height,
            width_factor,
            oblique_angle,
            tracking_percentage,
        ) = bs.read_struct("<4d")
        (
            is_backwards,
            is_upside_down,
            is_vertical,
            is_underline,
            is_overline,
        ) = bs.read_struct("<5L")
        font_filename = bs.read_padded_string()
        big_font_filename = bs.read_padded_string()
        attribs = self._build_dxf_attribs()
        attribs["insert"] = start_point
        attribs["text"] = text
        attribs["height"] = height
        attribs["width"] = width_factor
        attribs["rotation"] = text_direction.angle_deg
        attribs["oblique"] = math.degrees(oblique_angle)
        attribs["style"] = self._get_style(font_filename, big_font_filename)
        attribs["text_generation_flag"] = 2 * is_backwards + 4 * is_upside_down
        attribs["extrusion"] = normal
        return self._factory("TEXT", dxfattribs=attribs)

    def unicode_text2(self, data: bytes):
        bs = ByteStream(data)
        start_point = Vec3(bs.read_vertex())
        normal = Vec3(bs.read_vertex())
        text_direction = Vec3(bs.read_vertex())
        text = bs.read_padded_unicode_string()
        ignore_length_of_string, ignore_raw = bs.read_struct("<2l")
        (
            height,
            width_factor,
            oblique_angle,
            tracking_percentage,
        ) = bs.read_struct("<4d")
        (
            is_backwards,
            is_upside_down,
            is_vertical,
            is_underline,
            is_overline,
        ) = bs.read_struct("<5L")
        is_bold, is_italic, charset, pitch = bs.read_struct("<4L")
        type_face = bs.read_padded_unicode_string()
        font_filename = bs.read_padded_unicode_string()
        big_font_filename = bs.read_padded_unicode_string()
        attribs = self._build_dxf_attribs()
        attribs["insert"] = start_point
        attribs["text"] = text
        attribs["height"] = height
        attribs["width"] = width_factor
        attribs["rotation"] = text_direction.angle_deg
        attribs["oblique"] = math.degrees(oblique_angle)
        attribs["style"] = self._get_style(font_filename, big_font_filename)
        attribs["text_generation_flag"] = 2 * is_backwards + 4 * is_upside_down
        attribs["extrusion"] = normal
        return self._factory("TEXT", dxfattribs=attribs)

    def xline(self, data: bytes):
        return self._xline(data, "XLINE")

    def ray(self, data: bytes):
        return self._xline(data, "RAY")

    def _xline(self, data: bytes, type_: str):
        logger.warning(
            "Untested proxy graphic entity: RAY/XLINE - Need examples!"
        )
        bs = ByteStream(data)
        attribs = self._build_dxf_attribs()
        start_point = Vec3(bs.read_vertex())
        other_point = Vec3(bs.read_vertex())
        attribs["start"] = start_point
        attribs["unit_vector"] = (other_point - start_point).normalize()
        return self._factory(type_, dxfattribs=attribs)

    def _get_style(self, font: str, bigfont: str) -> str:
        self.required_fonts.add(font)
        if font in self.textstyles:
            style = self.textstyles[font]
        else:
            style = font
            if self._doc and not self._doc.styles.has_entry(style):
                self._doc.styles.new(
                    font, dxfattribs={"font": font, "bigfont": bigfont}
                )
                self.textstyles[font] = style
        return style

    def _load_vertices(self, data: bytes, load_normal=False):
        normal = Z_AXIS
        bs = ByteStream(data)
        count = bs.read_long()
        if load_normal:
            count += 1
        vertices = []
        while count > 0:
            vertices.append(Vec3(bs.read_struct("<3d")))
            count -= 1
        if load_normal:
            normal = vertices.pop()
        return vertices, normal

    def _build_dxf_attribs(self) -> Dict:
        attribs: Dict[str, Any] = dict()
        if self.layer != "0":
            attribs["layer"] = self.layer
        if self.color != const.BYLAYER:
            attribs["color"] = self.color
        if self.linetype != "BYLAYER":
            attribs["linetype"] = self.linetype
        if self.lineweight != const.LINEWEIGHT_DEFAULT:
            attribs["lineweight"] = self.lineweight
        if self.ltscale != 1.0:
            attribs["ltscale"] = self.ltscale
        if self.true_color is not None:
            attribs["true_color"] = self.true_color
        return attribs


class ProxyGraphicDebugger(ProxyGraphic):
    def __init__(self, data: bytes, doc: "Drawing" = None, debug_stream=None):
        super(ProxyGraphicDebugger, self).__init__(data, doc)
        if debug_stream is None:
            debug_stream = sys.stdout
        self._debug_stream = debug_stream

    def log_entities(self):
        self.log_separator(char="=", newline=False)
        self.log_message("Create virtual DXF entities:")
        self.log_separator(newline=False)
        for entity in self.virtual_entities():
            self.log_message(f"\n  * {entity.dxftype()}")
            self.log_message(f"  * {entity.graphic_properties()}\n")
        self.log_separator(char="=")

    def log_commands(self):
        self.log_separator(char="=", newline=False)
        self.log_message("Raw proxy commands:")
        self.log_separator(newline=False)
        for index, size, cmd in self.info():
            self.log_message(f"Command: {cmd} Index: {index} Size: {size}")
        self.log_separator(char="=")

    def log_separator(self, char="-", newline=True):
        self.log_message(char * 79)
        if newline:
            self.log_message("")

    def log_message(self, msg: str):
        print(msg, file=self._debug_stream)

    def log_state(self):
        self.log_message("> " + self.get_state())

    def get_state(self) -> str:
        return (
            f"ly: '{self.layer}', clr: {self.color}, lt: {self.linetype}, "
            f"lw: {self.lineweight}, ltscale: {self.ltscale}, "
            f"rgb: {self.true_color}, fill: {self.fill}"
        )

    def attribute_color(self, data: bytes):
        self.log_message("Command: set COLOR")
        super().attribute_color(data)
        self.log_state()

    def attribute_layer(self, data: bytes):
        self.log_message("Command: set LAYER")
        super().attribute_layer(data)
        self.log_state()

    def attribute_linetype(self, data: bytes):
        self.log_message("Command: set LINETYPE")
        super().attribute_linetype(data)
        self.log_state()

    def attribute_true_color(self, data: bytes):
        self.log_message("Command: set TRUE-COLOR")
        super().attribute_true_color(data)
        self.log_state()

    def attribute_lineweight(self, data: bytes):
        self.log_message("Command: set LINEWEIGHT")
        super().attribute_lineweight(data)
        self.log_state()

    def attribute_ltscale(self, data: bytes):
        self.log_message("Command: set LTSCALE")
        super().attribute_ltscale(data)
        self.log_state()

    def attribute_fill(self, data: bytes):
        self.log_message("Command: set FILL")
        super().attribute_fill(data)
        self.log_state()


def _apply_face_colors(polyface: "Polyface", colors: List[int]) -> None:
    color_count: int = len(colors)
    if color_count == 0:
        return

    index: int = 0
    for vertex in polyface.vertices:
        if vertex.is_face_record:
            vertex.dxf.color = colors[index]
            index += 1
            if index >= color_count:
                return



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from typing import (
    TYPE_CHECKING,
    Iterable,
    cast,
    Tuple,
    Union,
    Optional,
    Callable,
    Dict,
    List,
)
import math
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import (
    DXF12,
    SUBCLASS_MARKER,
    DXFValueError,
    DXFKeyError,
    DXFStructureError,
)
from ezdxf.math import (
    Vec3,
    X_AXIS,
    Y_AXIS,
    Z_AXIS,
    Matrix44,
    OCS,
    UCS,
    NULLVEC,
)
from ezdxf.math.transformtools import OCSTransform, InsertTransformationError
from ezdxf.explode import (
    explode_block_reference,
    virtual_block_reference_entities,
)
from ezdxf.entities import factory
from ezdxf.query import EntityQuery
from ezdxf.audit import AuditError
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity, elevation_to_z_axis
from .subentity import LinkedEntities
from .attrib import Attrib

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        TagWriter,
        Vertex,
        DXFNamespace,
        AttDef,
        BlockLayout,
        BaseLayout,
        Auditor,
    )

__all__ = ["Insert"]

ABS_TOL = 1e-9

# DXF files as XREF:
# The INSERT entity is used to attach XREFS.
# The model space is the block content, if the the whole document is used as
# BLOCK (XREF), but this is only supported for the DWG format.
# AutoCAD does not support DXF files as XREFS, they are ignored, but the DXF
# file is valid! BricsCAD shows DXF files as XREFS, but does not allow to attach
# DXF files as XREFS by the application itself.

# Multi-INSERT has subclass id AcDbMInsertBlock
acdb_block_reference = DefSubclass(
    "AcDbBlockReference",
    {
        "attribs_follow": DXFAttr(66, default=0, optional=True),
        "name": DXFAttr(2, validator=validator.is_valid_block_name),
        "insert": DXFAttr(10, xtype=XType.any_point),
        # Elevation is a legacy feature from R11 and prior, do not use this
        # attribute, store the entity elevation in the z-axis of the vertices.
        # ezdxf does not export the elevation attribute!
        "elevation": DXFAttr(38, default=0, optional=True),
        "xscale": DXFAttr(
            41,
            default=1,
            optional=True,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        "yscale": DXFAttr(
            42,
            default=1,
            optional=True,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        "zscale": DXFAttr(
            43,
            default=1,
            optional=True,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        "rotation": DXFAttr(50, default=0, optional=True),
        "column_count": DXFAttr(
            70,
            default=1,
            optional=True,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        "row_count": DXFAttr(
            71,
            default=1,
            optional=True,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        "column_spacing": DXFAttr(44, default=0, optional=True),
        "row_spacing": DXFAttr(45, default=0, optional=True),
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
    },
)
acdb_block_reference_group_codes = group_code_mapping(acdb_block_reference)
NON_ORTHO_MSG = (
    "INSERT entity can not represent a non-orthogonal target "
    "coordinate system."
)


# Notes to SEQEND:
#
# The INSERT entity requires only a SEQEND if ATTRIB entities are attached.
#  So a loaded INSERT could have a missing SEQEND.
#
# A bounded INSERT needs a SEQEND to be valid at export if there are attached
# ATTRIB entities, but the LinkedEntities.post_bind_hook() method creates
# always a new SEQEND after binding the INSERT entity to a document.
#
# Nonetheless the Insert.add_attrib() method also creates a requires SEQEND if
# necessary.


@factory.register_entity
class Insert(LinkedEntities):
    """DXF INSERT entity"""

    DXFTYPE = "INSERT"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_block_reference)

    @property
    def attribs(self) -> List[Attrib]:
        return self._sub_entities  # type: ignore

    @property
    def attribs_follow(self) -> bool:
        return bool(len(self.attribs))

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            # Always use the 2nd subclass, could be AcDbBlockReference or
            # AcDbMInsertBlock:
            processor.fast_load_dxfattribs(
                dxf, acdb_block_reference_group_codes, 2, recover=True
            )
            if processor.r12:
                # Transform elevation attribute from R11 to z-axis values:
                elevation_to_z_axis(dxf, ("insert",))
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            if (self.dxf.column_count > 1) or (self.dxf.row_count > 1):
                tagwriter.write_tag2(SUBCLASS_MARKER, "AcDbMInsertBlock")
            else:
                tagwriter.write_tag2(SUBCLASS_MARKER, "AcDbBlockReference")
        if self.attribs_follow:
            tagwriter.write_tag2(66, 1)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "insert",
                "xscale",
                "yscale",
                "zscale",
                "rotation",
                "column_count",
                "row_count",
                "column_spacing",
                "row_spacing",
                "extrusion",
            ],
        )

    def export_dxf(self, tagwriter: "TagWriter"):
        super().export_dxf(tagwriter)
        # Do no export SEQEND if no ATTRIBS attached:
        if self.attribs_follow:
            self.process_sub_entities(lambda e: e.export_dxf(tagwriter))

    @property
    def has_scaling(self) -> bool:
        """Returns ``True`` if any axis scaling is applied."""
        if self.dxf.hasattr("xscale") and self.dxf.xscale != 1:
            return True
        if self.dxf.hasattr("yscale") and self.dxf.yscale != 1:
            return True
        if self.dxf.hasattr("zscale") and self.dxf.zscale != 1:
            return True
        return False

    @property
    def has_uniform_scaling(self) -> bool:
        """Returns ``True`` if scaling is uniform in x-, y- and z-axis ignoring
        reflections e.g. (1, 1, -1) is uniform scaling.

        """
        return (
            abs(self.dxf.xscale) == abs(self.dxf.yscale) == abs(self.dxf.zscale)
        )

    def set_scale(self, factor: float):
        """Set uniform scaling."""
        if factor == 0:
            raise ValueError("Invalid scaling factor.")
        self.dxf.xscale = factor
        self.dxf.yscale = factor
        self.dxf.zscale = factor
        return self

    def is_xref(self) -> bool:
        """Return ``True`` if XREF or XREF_OVERLAY."""
        assert self.doc is not None, "Requires a document object"
        block_layout = self.doc.blocks.get(self.dxf.name)
        if (
            block_layout is not None and block_layout.block.dxf.flags & 12  # type: ignore
        ):  # XREF(4) & XREF_OVERLAY(8)
            return True
        return False

    def block(self) -> Optional["BlockLayout"]:
        """Returns associated :class:`~ezdxf.layouts.BlockLayout`."""
        if self.doc is None:
            return None
        return self.doc.blocks.get(self.dxf.name)

    def place(
        self,
        insert: "Vertex" = None,
        scale: Tuple[float, float, float] = None,
        rotation: float = None,
    ) -> "Insert":
        """
        Set block reference placing location `insert`, scaling and rotation
        attributes. Parameters which are ``None`` will not be altered.

        Args:
            insert: insert location as ``(x, y [,z])`` tuple
            scale: ``(x-scale, y-scale, z-scale)`` tuple
            rotation : rotation angle in degrees

        """
        if insert is not None:
            self.dxf.insert = insert
        if scale is not None:
            if len(scale) != 3:
                raise DXFValueError(
                    "Parameter scale has to be a (x, y, z)-tuple."
                )
            x, y, z = scale
            self.dxf.xscale = x
            self.dxf.yscale = y
            self.dxf.zscale = z
        if rotation is not None:
            self.dxf.rotation = rotation
        return self

    def grid(
        self,
        size: Tuple[int, int] = (1, 1),
        spacing: Tuple[float, float] = (1, 1),
    ) -> "Insert":
        """Place block reference in a grid layout, grid `size` defines the
        row- and column count, `spacing` defines the distance between two block
        references.

        Args:
            size: grid size as ``(row_count, column_count)`` tuple
            spacing: distance between placing as
                ``(row_spacing, column_spacing)`` tuple

        """
        try:
            rows, cols = size
        except ValueError:
            raise DXFValueError(
                "Size has to be a 2-tuple: (row_count, column_count)."
            )
        self.dxf.row_count = rows
        self.dxf.column_count = cols
        try:
            row_spacing, col_spacing = spacing
        except ValueError:
            raise DXFValueError(
                "Spacing has to be a 2-tuple: (row_spacing, column_spacing)."
            )
        self.dxf.row_spacing = row_spacing
        self.dxf.column_spacing = col_spacing
        return self

    def get_attrib(
        self, tag: str, search_const: bool = False
    ) -> Optional[Union["Attrib", "AttDef"]]:
        """Get attached :class:`Attrib` entity with :code:`dxf.tag == tag`,
        returns ``None`` if not found. Some applications may not attach constant
        ATTRIB entities, set `search_const` to ``True``, to get at least the
        associated :class:`AttDef` entity.

        Args:
            tag: tag name
            search_const: search also const ATTDEF entities

        """
        for attrib in self.attribs:
            if tag == attrib.dxf.tag:
                return attrib
        if search_const and self.doc is not None:
            block = self.doc.blocks[self.dxf.name]
            for attdef in block.get_const_attdefs():
                if tag == attdef.dxf.tag:
                    return attdef
        return None

    def get_attrib_text(
        self, tag: str, default: str = "", search_const: bool = False
    ) -> str:
        """Get content text of attached :class:`Attrib` entity with
        :code:`dxf.tag == tag`, returns `default` if not found.
        Some applications may not attach constant ATTRIB entities, set
        `search_const` to ``True``, to get content text of the
        associated :class:`AttDef` entity.

        Args:
            tag: tag name
            default: default value if ATTRIB `tag` is absent
            search_const: search also const ATTDEF entities

        """
        attrib = self.get_attrib(tag, search_const)
        if attrib is None:
            return default
        return attrib.dxf.text

    def has_attrib(self, tag: str, search_const: bool = False) -> bool:
        """Returns ``True`` if ATTRIB `tag` exist, for `search_const` doc see
        :meth:`get_attrib`.

        Args:
            tag: tag name as string
            search_const: search also const ATTDEF entities

        """
        return self.get_attrib(tag, search_const) is not None

    def add_attrib(
        self,
        tag: str,
        text: str,
        insert: "Vertex" = (0, 0),
        dxfattribs: dict = None,
    ) -> "Attrib":
        """Attach an :class:`Attrib` entity to the block reference.

        Example for appending an attribute to an INSERT entity with none
        standard alignment::

            e.add_attrib('EXAMPLETAG', 'example text').set_pos(
                (3, 7), align='MIDDLE_CENTER'
            )

        Args:
            tag: tag name as string
            text: content text as string
            insert: insert location as tuple ``(x, y[, z])`` in :ref:`WCS`
            dxfattribs: additional DXF attributes for the ATTRIB entity

        """
        dxfattribs = dxfattribs or {}
        dxfattribs["tag"] = tag
        dxfattribs["text"] = text
        dxfattribs["insert"] = insert
        attrib = cast("Attrib", self._new_compound_entity("ATTRIB", dxfattribs))
        self.attribs.append(attrib)

        # This case is only possible if INSERT is read from file without
        # attached ATTRIBS:
        if self.seqend is None:
            self.new_seqend()
        return attrib

    def delete_attrib(self, tag: str, ignore=False) -> None:
        """Delete an attached :class:`Attrib` entity from INSERT. If `ignore`
        is ``False``, an :class:`DXFKeyError` exception is raised, if
        ATTRIB `tag` does not exist.

        Args:
            tag: ATTRIB name
            ignore: ``False`` for raising :class:`DXFKeyError` if ATTRIB `tag`
                does not exist.

        Raises:
            DXFKeyError: if ATTRIB `tag` does not exist.

        """
        for index, attrib in enumerate(self.attribs):
            if attrib.dxf.tag == tag:
                del self.attribs[index]
                attrib.destroy()
                return
        if not ignore:
            raise DXFKeyError(tag)

    def delete_all_attribs(self) -> None:
        """Delete all :class:`Attrib` entities attached to the INSERT entity."""
        if not self.is_alive:
            return

        for attrib in self.attribs:
            attrib.destroy()
        self._sub_entities = []

    def transform(self, m: "Matrix44") -> "Insert":
        """Transform INSERT entity by transformation matrix `m` inplace.

        Unlike the transformation matrix `m`, the INSERT entity can not
        represent a non orthogonal target coordinate system, for this case an
        :class:`InsertTransformationError` will be raised.

        """

        dxf = self.dxf
        ocs = self.ocs()

        # Transform source OCS axis into the target coordinate system:
        ux, uy, uz = m.transform_directions((ocs.ux, ocs.uy, ocs.uz))

        # Calculate new axis scaling factors:
        x_scale = ux.magnitude * dxf.xscale
        y_scale = uy.magnitude * dxf.yscale
        z_scale = uz.magnitude * dxf.zscale

        ux = ux.normalize()
        uy = uy.normalize()
        uz = uz.normalize()
        # check for orthogonal x-, y- and z-axis
        if (
            abs(ux.dot(uz)) > ABS_TOL
            or abs(ux.dot(uy)) > ABS_TOL
            or abs(uz.dot(uy)) > ABS_TOL
        ):
            raise InsertTransformationError(NON_ORTHO_MSG)

        # expected y-axis for an orthogonal right handed coordinate system:
        expected_uy = uz.cross(ux)
        if not expected_uy.isclose(uy, abs_tol=ABS_TOL):
            # new y-axis points into opposite direction:
            y_scale = -y_scale

        ocs_transform = OCSTransform.from_ocs(OCS(dxf.extrusion), OCS(uz), m)
        dxf.insert = ocs_transform.transform_vertex(dxf.insert)
        dxf.rotation = ocs_transform.transform_deg_angle(dxf.rotation)

        dxf.extrusion = uz
        dxf.xscale = x_scale
        dxf.yscale = y_scale
        dxf.zscale = z_scale

        for attrib in self.attribs:
            attrib.transform(m)
        self.post_transform(m)
        return self

    def translate(self, dx: float, dy: float, dz: float) -> "Insert":
        """Optimized INSERT translation about `dx` in x-axis, `dy` in y-axis
        and `dz` in z-axis.

        """
        ocs = self.ocs()
        self.dxf.insert = ocs.from_wcs(
            Vec3(dx, dy, dz) + ocs.to_wcs(self.dxf.insert)
        )
        for attrib in self.attribs:
            attrib.translate(dx, dy, dz)
        return self

    def matrix44(self) -> Matrix44:
        """Returns a transformation matrix of type :class:`Matrix44` to
        transform the block entities into :ref:`WCS`.

        """
        dxf = self.dxf
        sx = dxf.xscale
        sy = dxf.yscale
        sz = dxf.zscale

        ocs = self.ocs()
        extrusion = ocs.uz
        ux = Vec3(ocs.to_wcs(X_AXIS))
        uy = Vec3(ocs.to_wcs(Y_AXIS))
        m = Matrix44.ucs(ux=ux * sx, uy=uy * sy, uz=extrusion * sz)
        # same as Matrix44.ucs(ux, uy, extrusion) * Matrix44.scale(sx, sy, sz)

        angle = math.radians(dxf.rotation)
        if angle:
            m *= Matrix44.axis_rotate(extrusion, angle)

        insert = ocs.to_wcs(dxf.get("insert", NULLVEC))

        block_layout = self.block()
        if block_layout is not None:
            # transform block base point into WCS without translation
            insert -= m.transform_direction(block_layout.block.dxf.base_point)  # type: ignore

        # set translation
        m.set_row(3, insert.xyz)  # type: ignore
        return m

    def ucs(self):
        """Returns the block reference coordinate system as
        :class:`ezdxf.math.UCS` object.
        """
        m = self.matrix44()
        ucs = UCS()
        ucs.matrix = m
        return ucs

    def reset_transformation(self) -> None:
        """Reset block reference parameters `location`, `rotation` and
        `extrusion` vector.

        """
        self.dxf.insert = NULLVEC
        self.dxf.discard("rotation")
        self.dxf.discard("extrusion")

    def explode(self, target_layout: "BaseLayout" = None) -> "EntityQuery":
        """Explode block reference entities into target layout, if target
        layout is ``None``, the target layout is the layout of the block
        reference. This method destroys the source block reference entity.

        Transforms the block entities into the required :ref:`WCS` location by
        applying the block reference attributes `insert`, `extrusion`,
        `rotation` and the scaling values `xscale`, `yscale` and `zscale`.

        Attached ATTRIB entities are converted to TEXT entities, this is the
        behavior of the BURST command of the AutoCAD Express Tools.

        Returns an :class:`~ezdxf.query.EntityQuery` container with all
        "exploded" DXF entities.

        .. warning::

            **Non uniform scaling** may lead to incorrect results for text
            entities (TEXT, MTEXT, ATTRIB) and maybe some other entities.

        Args:
            target_layout: target layout for exploded entities, ``None`` for
                same layout as source entity.

        """
        if target_layout is None:
            target_layout = self.get_layout()
            if target_layout is None:
                raise DXFStructureError(
                    "INSERT without layout assignment, specify target layout."
                )
        return explode_block_reference(self, target_layout=target_layout)

    def __virtual_entities__(self) -> Iterable[DXFGraphic]:
        """Implements the SupportsVirtualEntities protocol.

        This protocol is for consistent internal usage and does not replace
        the method :meth:`virtual_entities`!
        """
        return virtual_block_reference_entities(self)

    def virtual_entities(
        self,
        skipped_entity_callback: Optional[
            Callable[[DXFGraphic, str], None]
        ] = None,
    ) -> Iterable[DXFGraphic]:
        """
        Yields "virtual" entities of a block reference. This method is meant to
        examine the block reference entities at the "exploded" location without
        really "exploding" the block reference. The`skipped_entity_callback()`
        will be called for all entities which are not processed, signature:
        :code:`skipped_entity_callback(entity: DXFEntity, reason: str)`,
        `entity` is the original (untransformed) DXF entity of the block
        definition, the `reason` string is an explanation why the entity was
        skipped.

        This entities are not stored in the entity database, have no handle and
        are not assigned to any layout. It is possible to convert this entities
        into regular drawing entities by adding the entities to the entities
        database and a layout of the same DXF document as the block reference::

            doc.entitydb.add(entity)
            msp = doc.modelspace()
            msp.add_entity(entity)

        This method does not resolve the MINSERT attributes, only the
        sub-entities of the base INSERT will be returned. To resolve MINSERT
        entities check if multi insert processing is required, that's the case
        if property :attr:`Insert.mcount` > 1, use the :meth:`Insert.multi_insert`
        method to resolve the MINSERT entity into single INSERT entities.

        .. warning::

            **Non uniform scaling** may return incorrect results for text
            entities (TEXT, MTEXT, ATTRIB) and maybe some other entities.

        Args:
            skipped_entity_callback: called whenever the transformation of an
                entity is not supported and so was skipped

        """
        return virtual_block_reference_entities(
            self, skipped_entity_callback=skipped_entity_callback
        )

    @property
    def mcount(self):
        """Returns the multi-insert count, MINSERT (multi-insert) processing
        is required if :attr:`mcount` > 1.

        .. versionadded:: 0.14

        """
        return (self.dxf.row_count if self.dxf.row_spacing else 1) * (
            self.dxf.column_count if self.dxf.column_spacing else 1
        )

    def multi_insert(self) -> Iterable["Insert"]:
        """Yields a virtual INSERT entity for each grid element of a MINSERT
        entity (multi-insert).

        .. versionadded:: 0.14

        """

        def transform_attached_attrib_entities(insert, offset):
            for attrib in insert.attribs:
                attrib.dxf.insert += offset

        def adjust_dxf_attribs(insert, offset):
            dxf = insert.dxf
            dxf.insert += offset
            dxf.discard("row_count")
            dxf.discard("column_count")
            dxf.discard("row_spacing")
            dxf.discard("column_spacing")

        done = set()
        row_spacing = self.dxf.row_spacing
        col_spacing = self.dxf.column_spacing
        rotation = self.dxf.rotation
        for row in range(self.dxf.row_count):
            for col in range(self.dxf.column_count):
                # All transformations in OCS:
                offset = Vec3(col * col_spacing, row * row_spacing)
                # If any spacing is 0, yield only unique locations:
                if offset not in done:
                    done.add(offset)
                    if rotation:  # Apply rotation to the grid.
                        offset = offset.rotate_deg(rotation)
                    # Do not apply scaling to the grid!
                    insert = self.copy()
                    adjust_dxf_attribs(insert, offset)
                    transform_attached_attrib_entities(insert, offset)
                    yield insert

    def add_auto_attribs(self, values: Dict[str, str]) -> "Insert":
        """
        Attach for each :class:`~ezdxf.entities.Attdef` entity, defined in the
        block definition, automatically an :class:`Attrib` entity to the block
        reference and set ``tag/value`` DXF attributes of the ATTRIB entities
        by the ``key/value`` pairs (both as strings) of the `values` dict.
        The ATTRIB entities are placed relative to the insert location of the
        block reference, which is identical to the block base point.

        This method avoids the wrapper block of the
        :meth:`~ezdxf.layouts.BaseLayout.add_auto_blockref` method, but the
        visual results may not match the results of CAD applications, especially
        for non uniform scaling. If the visual result is very important to you,
        use the :meth:`add_auto_blockref` method.

        Args:
            values: :class:`~ezdxf.entities.Attrib` tag values as ``tag/value``
                pairs

        """

        def unpack(dxfattribs) -> Tuple[str, str, "Vertex"]:
            tag = dxfattribs.pop("tag")
            text = values.get(tag, "")
            location = dxfattribs.pop("insert")
            return tag, text, location

        def autofill() -> None:
            for attdef in blockdef.attdefs():  # type: ignore
                dxfattribs = attdef.dxfattribs(drop={"prompt", "handle"})
                tag, text, location = unpack(dxfattribs)
                attrib = self.add_attrib(tag, text, location, dxfattribs)
                attrib.transform(m)

        m = self.matrix44()
        blockdef = self.block()
        autofill()
        return self

    def audit(self, auditor: "Auditor") -> None:
        """Validity check."""
        super().audit(auditor)
        doc = auditor.doc
        if doc and doc.blocks:
            if self.dxf.name not in doc.blocks:
                auditor.fixed_error(
                    code=AuditError.UNDEFINED_BLOCK,
                    message=f"Deleted entity {str(self)} without required BLOCK"
                            f" definition.",
                )
                auditor.trash(self)



# Purpose: read and write AutoCAD CTB files
# Copyright (c) 2010-2021, Manfred Moitzi
# License: MIT License
# IMPORTANT: use only standard 7-Bit ascii code

from typing import (
    Union,
    Tuple,
    Optional,
    BinaryIO,
    TextIO,
    Iterable,
    List,
    Any,
    Dict,
)
from abc import abstractmethod
from io import StringIO
from array import array
from struct import pack
import zlib

END_STYLE_BUTT = 0
END_STYLE_SQUARE = 1
END_STYLE_ROUND = 2
END_STYLE_DIAMOND = 3
END_STYLE_OBJECT = 4

JOIN_STYLE_MITER = 0
JOIN_STYLE_BEVEL = 1
JOIN_STYLE_ROUND = 2
JOIN_STYLE_DIAMOND = 3
JOIN_STYLE_OBJECT = 5

FILL_STYLE_SOLID = 64
FILL_STYLE_CHECKERBOARD = 65
FILL_STYLE_CROSSHATCH = 66
FILL_STYLE_DIAMONDS = 67
FILL_STYLE_HORIZONTAL_BARS = 68
FILL_STYLE_SLANT_LEFT = 69
FILL_STYLE_SLANT_RIGHT = 70
FILL_STYLE_SQUARE_DOTS = 71
FILL_STYLE_VERICAL_BARS = 72
FILL_STYLE_OBJECT = 73

DITHERING_ON = 1  # bit coded color_policy
GRAYSCALE_ON = 2  # bit coded color_policy
NAMED_COLOR = 4  # bit coded color_policy

AUTOMATIC = 0
OBJECT_LINEWEIGHT = 0
OBJECT_LINETYPE = 31
OBJECT_COLOR = -1
OBJECT_COLOR2 = -1006632961

STYLE_COUNT = 255

DEFAULT_LINE_WEIGHTS = [
    0.00,  # 0
    0.05,  # 1
    0.09,  # 2
    0.10,  # 3
    0.13,  # 4
    0.15,  # 5
    0.18,  # 6
    0.20,  # 7
    0.25,  # 8
    0.30,  # 9
    0.35,  # 10
    0.40,  # 11
    0.45,  # 12
    0.50,  # 13
    0.53,  # 14
    0.60,  # 15
    0.65,  # 16
    0.70,  # 17
    0.80,  # 18
    0.90,  # 19
    1.00,  # 20
    1.06,  # 21
    1.20,  # 22
    1.40,  # 23
    1.58,  # 24
    2.00,  # 25
    2.11,  # 26
]

# color_type: (thx to Rammi)

# Take color from layer, ignore other bytes.
COLOR_BY_LAYER = 0xC0

# Take color from insertion, ignore other bytes
COLOR_BY_BLOCK = 0xC1

# RGB value, other bytes are R,G,B.
COLOR_RGB = 0xC2

# ACI, AutoCAD color index, other bytes are 0,0,index ???
COLOR_ACI = 0xC3


def color_name(index: int) -> str:
    return "Color_%d" % (index + 1)


def get_bool(value: Union[str, bool]) -> bool:
    if isinstance(value, str):
        upperstr = value.upper()
        if upperstr == "TRUE":
            value = True
        elif upperstr == "FALSE":
            value = False
        else:
            raise ValueError("Unknown bool value '%s'." % str(value))
    return value


class PlotStyle:
    def __init__(
        self, index: int, data: dict = None, parent: "PlotStyleTable" = None
    ):
        data = data or {}
        self.parent = parent
        self.index = int(index)
        self.name = str(data.get("name", color_name(index)))
        self.localized_name = str(data.get("localized_name", color_name(index)))
        self.description = str(data.get("description", ""))
        # do not set _color, _mode_color or _color_policy directly
        # use set_color() method, and the properties dithering and grayscale
        self._color = int(data.get("color", OBJECT_COLOR))
        self._color_type = COLOR_RGB
        if self._color != OBJECT_COLOR:
            self._mode_color = int(data.get("mode_color", self._color))
        self._color_policy = int(data.get("color_policy", DITHERING_ON))
        self.physical_pen_number = int(
            data.get("physical_pen_number", AUTOMATIC)
        )
        self.virtual_pen_number = int(data.get("virtual_pen_number", AUTOMATIC))
        self.screen = int(data.get("screen", 100))
        self.linepattern_size = float(data.get("linepattern_size", 0.5))
        self.linetype = int(data.get("linetype", OBJECT_LINETYPE))  # 0 .. 30
        self.adaptive_linetype = get_bool(data.get("adaptive_linetype", True))

        # lineweight index
        self.lineweight = int(data.get("lineweight", OBJECT_LINEWEIGHT))
        self.end_style = int(data.get("end_style", END_STYLE_OBJECT))
        self.join_style = int(data.get("join_style", JOIN_STYLE_OBJECT))
        self.fill_style = int(data.get("fill_style", FILL_STYLE_OBJECT))

    @property
    def color(self) -> Optional[Tuple[int, int, int]]:
        """Get style color as ``(r, g, b)`` tuple or ``None``, if style has
        object color.
        """
        if self.has_object_color():
            return None  # object color
        else:
            return int2color(self._mode_color)[:3]

    @color.setter
    def color(self, rgb: Tuple[int, int, int]) -> None:
        """Set color as RGB values."""
        r, g, b = rgb
        # when defining a user-color, `mode_color` represents the real
        # true_color as (r, g, b) tuple and color_type = COLOR_RGB (0xC2) as
        # highest byte, the `color` value calculated for a user-color is not a
        # (r, g, b) tuple and has color_type = COLOR_ACI (0xC3) (sometimes), set
        # for `color` the same value as for `mode_color`, because AutoCAD
        # corrects the `color` value by itself.
        self._mode_color = mode_color2int(r, g, b, color_type=self._color_type)
        self._color = self._mode_color

    @property
    def color_type(self):
        if self.has_object_color():
            return None  # object color
        else:
            return self._color_type

    @color_type.setter
    def color_type(self, value: int):
        self._color_type = value

    def set_object_color(self) -> None:
        """Set color to object color."""
        self._color = OBJECT_COLOR
        self._mode_color = OBJECT_COLOR

    def set_lineweight(self, lineweight: float) -> None:
        """Set `lineweight` in millimeters. Use ``0.0`` to set lineweight by
        object.
        """
        assert self.parent is not None
        self.lineweight = self.parent.get_lineweight_index(lineweight)

    def get_lineweight(self) -> float:
        """Returns the lineweight in millimeters or `0.0` for use entity
        lineweight.
        """
        assert self.parent is not None
        return self.parent.lineweights[self.lineweight]

    def has_object_color(self) -> bool:
        """``True`` if style has object color."""
        return self._color in (OBJECT_COLOR, OBJECT_COLOR2)

    @property
    def aci(self) -> int:
        """:ref:`ACI` in range from ``1`` to ``255``. Has no meaning for named
        plot styles. (int)
        """
        return self.index + 1

    @property
    def dithering(self) -> bool:
        """Depending on the capabilities of your plotter, dithering approximates
        the colors with dot patterns. When this option is ``False``, the colors
        are mapped to the nearest color, resulting in a smaller range of
        colors when plotting.

        Dithering is available only whether you select the object’s color or
        assign a plot style color.

        """
        return bool(self._color_policy & DITHERING_ON)

    @dithering.setter
    def dithering(self, status: bool) -> None:
        if status:
            self._color_policy |= DITHERING_ON
        else:
            self._color_policy &= ~DITHERING_ON

    @property
    def grayscale(self) -> bool:
        """Plot colors in grayscale. (bool)"""
        return bool(self._color_policy & GRAYSCALE_ON)

    @grayscale.setter
    def grayscale(self, status: bool) -> None:
        if status:
            self._color_policy |= GRAYSCALE_ON
        else:
            self._color_policy &= ~GRAYSCALE_ON

    @property
    def named_color(self) -> bool:
        return bool(self._color_policy & NAMED_COLOR)

    @named_color.setter
    def named_color(self, status: bool) -> None:
        if status:
            self._color_policy |= NAMED_COLOR
        else:
            self._color_policy &= ~NAMED_COLOR

    def write(self, stream: TextIO) -> None:
        """Write style data to file-like object `stream`."""
        index = self.index
        stream.write(" %d{\n" % index)
        stream.write('  name="%s\n' % self.name)
        stream.write('  localized_name="%s\n' % self.localized_name)
        stream.write('  description="%s\n' % self.description)
        stream.write("  color=%d\n" % self._color)
        if self._color != OBJECT_COLOR:
            stream.write("  mode_color=%d\n" % self._mode_color)
        stream.write("  color_policy=%d\n" % self._color_policy)
        stream.write("  physical_pen_number=%d\n" % self.physical_pen_number)
        stream.write("  virtual_pen_number=%d\n" % self.virtual_pen_number)
        stream.write("  screen=%d\n" % self.screen)
        stream.write("  linepattern_size=%s\n" % str(self.linepattern_size))
        stream.write("  linetype=%d\n" % self.linetype)
        stream.write(
            "  adaptive_linetype=%s\n"
            % str(bool(self.adaptive_linetype)).upper()
        )
        stream.write("  lineweight=%s\n" % str(self.lineweight))
        stream.write("  fill_style=%d\n" % self.fill_style)
        stream.write("  end_style=%d\n" % self.end_style)
        stream.write("  join_style=%d\n" % self.join_style)
        stream.write(" }\n")


class PlotStyleTable:
    """PlotStyle container"""

    def __init__(
        self,
        description: str = "",
        scale_factor: float = 1.0,
        apply_factor: bool = False,
    ):
        self.description = description
        self.scale_factor = scale_factor
        self.apply_factor = apply_factor

        # set custom_lineweight_display_units to 1 for showing lineweight in inch in AutoCAD CTB editor window, but
        # lineweight is always defined in mm
        self.custom_lineweight_display_units = 0
        self.lineweights = array("f", DEFAULT_LINE_WEIGHTS)

    def get_lineweight_index(self, lineweight: float) -> int:
        """Get index of `lineweight` in the lineweight table or append
        `lineweight` to lineweight table.
        """
        try:
            return self.lineweights.index(lineweight)
        except ValueError:
            self.lineweights.append(lineweight)
            return len(self.lineweights) - 1

    def set_table_lineweight(self, index: int, lineweight: float) -> int:
        """Argument `index` is the lineweight table index, not the :ref:`ACI`.

        Args:
            index: lineweight table index = :attr:`PlotStyle.lineweight`
            lineweight: in millimeters

        """
        try:
            self.lineweights[index] = lineweight
            return index
        except IndexError:
            self.lineweights.append(lineweight)
            return len(self.lineweights) - 1

    def get_table_lineweight(self, index: int) -> float:
        """Returns lineweight in millimeters of lineweight table entry `index`.

        Args:
            index: lineweight table index = :attr:`PlotStyle.lineweight`

        Returns:
            lineweight in mm or ``0.0`` for use entity lineweight

        """
        return self.lineweights[index]

    def save(self, filename: str) -> None:
        """Save CTB or STB file as `filename` to the file system."""
        with open(filename, "wb") as stream:
            self.write(stream)

    def write(self, stream: BinaryIO) -> None:
        """Compress and write the CTB or STB file to binary `stream`."""
        memfile = StringIO()
        self.write_content(memfile)
        memfile.write(chr(0))  # end of file
        body = memfile.getvalue()
        memfile.close()
        _compress(stream, body)

    @abstractmethod
    def write_content(self, stream: TextIO) -> None:
        pass

    def _write_lineweights(self, stream: TextIO) -> None:
        """Write custom lineweight table to text `stream`."""
        stream.write("custom_lineweight_table{\n")
        for index, weight in enumerate(self.lineweights):
            stream.write(" %d=%.2f\n" % (index, weight))
        stream.write("}\n")

    def parse(self, text: str) -> None:
        """Parse plot styles from CTB string `text`."""

        def set_lineweights(lineweights):
            if lineweights is None:
                return
            self.lineweights = array("f", [0.0] * len(lineweights))
            for key, value in lineweights.items():
                self.lineweights[int(key)] = float(value)

        parser = PlotStyleFileParser(text)
        self.description = parser.get("description", "")
        self.scale_factor = float(parser.get("scale_factor", 1.0))
        self.apply_factor = get_bool(parser.get("apply_factor", True))
        self.custom_lineweight_display_units = int(
            parser.get("custom_lineweight_display_units", 0)
        )
        set_lineweights(parser.get("custom_lineweight_table", None))
        self.load_styles(parser.get("plot_style", {}))

    @abstractmethod
    def load_styles(self, styles):
        pass


class ColorDependentPlotStyles(PlotStyleTable):
    def __init__(
        self,
        description: str = "",
        scale_factor: float = 1.0,
        apply_factor: bool = False,
    ):
        super().__init__(description, scale_factor, apply_factor)
        self._styles = [
            PlotStyle(index, parent=self) for index in range(STYLE_COUNT)
        ]  # type: List[PlotStyle]
        self._styles.insert(
            0, PlotStyle(256)
        )  # 1-based array: insert dummy value for index 0

    def __getitem__(self, aci: int) -> PlotStyle:
        """Returns :class:`PlotStyle` for :ref:`ACI` `aci`."""
        if 0 < aci < 256:
            return self._styles[aci]
        else:
            raise IndexError(aci)

    def __setitem__(self, aci: int, style: PlotStyle):
        """Set plot `style` for `aci`."""
        if 0 < aci < 256:
            style.parent = self
            self._styles[aci] = style
        else:
            raise IndexError(aci)

    def __iter__(self):
        """Iterable of all plot styles."""
        return iter(self._styles[1:])

    def new_style(self, aci: int, data: dict = None) -> PlotStyle:
        """Set `aci` to new attributes defined by `data` dict.

        Args:
            aci: :ref:`ACI`
            data: ``dict`` of :class:`PlotStyle` attributes: description, color,
                physical_pen_number, virtual_pen_number, screen,
                linepattern_size, linetype, adaptive_linetype,
                lineweight, end_style, join_style, fill_style

        """
        # ctb table index = aci - 1
        # ctb table starts with index 0, where aci == 0 means BYBLOCK
        style = PlotStyle(index=aci - 1, data=data)
        style.color_type = COLOR_RGB
        self[aci] = style
        return style

    def get_lineweight(self, aci: int):
        """Returns the assigned lineweight for :class:`PlotStyle` `aci` in
        millimeter.
        """
        style = self[aci]
        lineweight = style.get_lineweight()
        if lineweight == 0.0:
            return None
        else:
            return lineweight

    def write_content(self, stream: TextIO) -> None:
        """Write the CTB-file to text `stream`."""
        self._write_header(stream)
        self._write_aci_table(stream)
        self._write_plot_styles(stream)
        self._write_lineweights(stream)

    def _write_header(self, stream: TextIO) -> None:
        """Write header values of CTB-file to text `stream`."""
        stream.write('description="%s\n' % self.description)
        stream.write("aci_table_available=TRUE\n")
        stream.write("scale_factor=%.1f\n" % self.scale_factor)
        stream.write("apply_factor=%s\n" % str(self.apply_factor).upper())
        stream.write(
            "custom_lineweight_display_units=%s\n"
            % str(self.custom_lineweight_display_units)
        )

    def _write_aci_table(self, stream: TextIO) -> None:
        """Write AutoCAD Color Index table to text `stream`."""
        stream.write("aci_table{\n")
        for style in self:
            index = style.index
            stream.write(' %d="%s\n' % (index, color_name(index)))
        stream.write("}\n")

    def _write_plot_styles(self, stream: TextIO) -> None:
        """Write user styles to text `stream`."""
        stream.write("plot_style{\n")
        for style in self:
            style.write(stream)
        stream.write("}\n")

    def load_styles(self, styles):
        for index, style in styles.items():
            index = int(index)
            style = PlotStyle(index, style)
            style.color_type = COLOR_RGB
            aci = index + 1
            self[aci] = style


class NamedPlotStyles(PlotStyleTable):
    def __init__(
        self,
        description: str = "",
        scale_factor: float = 1.0,
        apply_factor: bool = False,
    ):
        super().__init__(description, scale_factor, apply_factor)
        normal = PlotStyle(
            0,
            data={
                "name": "Normal",
                "localized_name": "Normal",
            },
        )
        self._styles = {"Normal": normal}  # type: Dict[str, PlotStyle]

    def __iter__(self) -> Iterable[str]:
        """Iterable of all plot style names."""
        return self.keys()

    def __getitem__(self, name: str) -> PlotStyle:
        """Returns :class:`PlotStyle` by `name`."""
        return self._styles[name]

    def __delitem__(self, name: str) -> None:
        """Delete plot style `name`. Plot style ``'Normal'`` is not deletable."""
        if name != "Normal":
            del self._styles[name]
        else:
            raise ValueError("Can't delete plot style 'Normal'. ")

    def keys(self) -> Iterable[str]:
        """Iterable of all plot style names."""
        keys = set(self._styles.keys())
        keys.discard("Normal")
        result = ["Normal"]
        result.extend(sorted(keys))
        return iter(result)

    def items(self) -> Iterable[Tuple[str, PlotStyle]]:
        """Iterable of all plot styles as (``name``, class:`PlotStyle`) tuples."""
        for key in self.keys():
            yield key, self._styles[key]

    def values(self) -> Iterable[PlotStyle]:
        """Iterable of all class:`PlotStyle` objects."""
        for key, value in self.items():
            yield value

    def new_style(
        self, name: str, data: dict = None, localized_name: str = None
    ) -> PlotStyle:
        """Create new class:`PlotStyle` `name` by attribute dict `data`, replaces
        existing class:`PlotStyle` objects.

        Args:
            name: plot style name
            localized_name: name shown in plot style editor, uses `name` if ``None``
            data: ``dict`` of :class:`PlotStyle` attributes: description, color,
                physical_pen_number, virtual_pen_number, screen,
                linepattern_size, linetype, adaptive_linetype, lineweight,
                end_style, join_style, fill_style

        """
        if name.lower() == "Normal":
            raise ValueError("Can't replace or modify plot style 'Normal'. ")
        data = data or {}
        data["name"] = name
        data["localized_name"] = localized_name or name
        index = len(self._styles)
        style = PlotStyle(index=index, data=data, parent=self)
        style.color_type = COLOR_ACI
        style.named_color = True
        self._styles[name] = style
        return style

    def get_lineweight(self, name: str):
        """Returns the assigned lineweight for :class:`PlotStyle` `name` in
        millimeter.
        """
        style = self[name]
        lineweight = style.get_lineweight()
        if lineweight == 0.0:
            return None
        else:
            return lineweight

    def write_content(self, stream: TextIO) -> None:
        """Write the STB-file to text `stream`."""
        self._write_header(stream)
        self._write_plot_styles(stream)
        self._write_lineweights(stream)

    def _write_header(self, stream: TextIO) -> None:
        """Write header values of CTB-file to text `stream`."""
        stream.write('description="%s\n' % self.description)
        stream.write("aci_table_available=FALSE\n")
        stream.write("scale_factor=%.1f\n" % self.scale_factor)
        stream.write("apply_factor=%s\n" % str(self.apply_factor).upper())
        stream.write(
            "custom_lineweight_display_units=%s\n"
            % str(self.custom_lineweight_display_units)
        )

    def _write_plot_styles(self, stream: TextIO) -> None:
        """Write user styles to text `stream`."""
        stream.write("plot_style{\n")
        for index, style in enumerate(self.values()):
            style.index = index
            style.write(stream)
        stream.write("}\n")

    def load_styles(self, styles):
        for index, style in styles.items():
            index = int(index)
            style = PlotStyle(index, style)
            style.color_type = COLOR_ACI
            self._styles[style.name] = style


def _read_ctb(stream: BinaryIO) -> ColorDependentPlotStyles:
    """Read a CTB-file from from binary `stream`."""
    content: bytes = _decompress(stream)
    styles = ColorDependentPlotStyles()
    styles.parse(content.decode())
    return styles


def _read_stb(stream: BinaryIO) -> NamedPlotStyles:
    """Read a STB-file from from binary `stream`."""
    content: bytes = _decompress(stream)
    styles = NamedPlotStyles()
    styles.parse(content.decode())
    return styles


def load(filename: str) -> Union[ColorDependentPlotStyles, NamedPlotStyles]:
    """Load the CTB or STB file `filename` from file system."""

    with open(filename, "rb") as stream:
        if filename.lower().endswith(".ctb"):
            return _read_ctb(stream)
        elif filename.lower().endswith(".stb"):
            return _read_stb(stream)
        else:
            raise ValueError('Invalid file type: "{}"'.format(filename))


def new_ctb() -> ColorDependentPlotStyles:
    """
    Create a new CTB file.

    .. versionchanged:: 0.10

        renamed from :func:`new`

    """
    return ColorDependentPlotStyles()


def new_stb() -> NamedPlotStyles:
    """Create a new STB file."""
    return NamedPlotStyles()


def _decompress(stream: BinaryIO) -> bytes:
    """Read and decompress the file content from binray `stream`."""
    content = stream.read()
    data = zlib.decompress(content[60:])  # type: bytes
    return data[:-1]  # truncate trailing \nul


def _compress(stream: BinaryIO, content: str):
    """Compress `content` and write to binary `stream`."""
    comp_body = zlib.compress(content.encode())
    adler_chksum = zlib.adler32(comp_body)
    stream.write(b"PIAFILEVERSION_2.0,CTBVER1,compress\r\npmzlibcodec")
    stream.write(pack("LLL", adler_chksum, len(content), len(comp_body)))
    stream.write(comp_body)


class PlotStyleFileParser:
    """A very simple CTB/STB file parser. CTB/STB files are created by
    applications, so the file structure should be correct in the most cases.
    """

    def __init__(self, text: str):
        """
        :param str text: ctb content as string

        """
        self.data = {}
        for element, value in PlotStyleFileParser.iteritems(text):
            self.data[element] = value

    @staticmethod
    def iteritems(text: str):
        """Iterate over all first level (start at col 0) elements."""
        line_index = 0

        def get_name() -> str:
            """Get element name of line <line_index>."""
            line = lines[line_index]
            if line.endswith("{"):  # start of a list like 'plot_style{'
                name = line[:-1]
            else:  # simple name=value line
                name = line.split("=", 1)[0]
            return name.strip()

        def get_mapping() -> dict:
            """Get mapping of elements enclosed by { }.

            e. g. lineweights, plot_styles, aci_table

            """

            def end_of_list():
                return lines[line_index].endswith("}")

            nonlocal line_index
            data = dict()
            while not end_of_list():
                name = get_name()
                value = get_value()  # get value or sub-list
                data[name] = value
            line_index += 1
            return data  # skip '}' - end of list

        def get_value() -> Union[str, dict]:
            """Get value of line <line_index> or the list that starts in line
            <line_index>.
            """
            nonlocal line_index
            line = lines[line_index]
            if line.endswith("{"):  # start of a list
                line_index += 1
                return get_mapping()
            else:  # it's a simple name=value line
                value: str = line.split("=", 1)[1]
                value = value.lstrip('"')  # strings look like this: name="value
                line_index += 1
            return value

        def skip_empty_lines():
            nonlocal line_index
            while line_index < len(lines) and len(lines[line_index]) == 0:
                line_index += 1

        lines = text.split("\n")
        while line_index < len(lines):
            name = get_name()
            value = get_value()
            yield (name, value)
            skip_empty_lines()

    def get(self, name: str, default: Any) -> Any:
        return self.data.get(name, default)


def int2color(color: int) -> Tuple[int, int, int, int]:
    """Convert color integer value from CTB-file to ``(r, g, b, color_type)
    tuple.
    """
    # Take color from layer, ignore other bytes.
    color_type = (color & 0xFF000000) >> 24
    red = (color & 0xFF0000) >> 16
    green = (color & 0xFF00) >> 8
    blue = color & 0xFF
    return red, green, blue, color_type


def mode_color2int(
    red: int, green: int, blue: int, color_type=COLOR_RGB
) -> int:
    """Convert mode_color (r, g, b, color_type) tuple to integer."""
    return -color2int(red, green, blue, color_type)


def color2int(red: int, green: int, blue: int, color_type: int) -> int:
    """Convert color (r, g, b, color_type) to integer."""
    return (
        -((color_type << 24) + (red << 16) + (green << 8) + blue) & 0xFFFFFFFF
    )



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import typing
from PyQt5.QtWidgets import QTableView, QTreeView
from PyQt5.QtCore import QModelIndex

if typing.TYPE_CHECKING:
    from ezdxf.eztypes import Tags


class StructureTree(QTreeView):
    def set_structure(self, model):
        self.setModel(model)
        self.expand(model.index(0, 0, QModelIndex()))
        self.setHeaderHidden(True)

    def expand_to_entity(self, entity: "Tags"):
        model = self.model()
        index = model.index_of_entity(entity)
        self.setCurrentIndex(index)


class DXFTagsTable(QTableView):
    def __init__(self):
        super().__init__()
        col_header = self.horizontalHeader()
        col_header.setStretchLastSection(True)
        row_header = self.verticalHeader()
        row_header.setDefaultSectionSize(24)  # default row height in pixels
        self.setSelectionBehavior(QTableView.SelectRows)

    def first_selected_row(self) -> int:
        first_row: int = 0
        selection = self.selectedIndexes()
        if selection:
            first_row = selection[0].row()
        return first_row



# Copyright (c) 2018-2021, Manfred Moitzi
# License: MIT License
from .arrows import ARROWS
from .r12spline import R12Spline
from .curves import Bezier, EulerSpiral, Spline, random_2d_path, random_3d_path
from .mesh import MeshBuilder, MeshVertexMerger, MeshTransformer, MeshAverageVertexMerger
from .trace import TraceBuilder



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import copy
import math
from typing import Iterable, List, Optional, Tuple

from ezdxf import colors
from ezdxf.entities import MText
from ezdxf.lldxf import const
from ezdxf.math import Matrix44, Vec3
from ezdxf.render.abstract_mtext_renderer import AbstractMTextRenderer
from ezdxf.tools import text_layout as tl, fonts
from ezdxf.tools.text import MTextContext
from .backend import BackendInterface
from .properties import Properties, RenderContext, rgb_to_hex
from .type_hints import Color

__all__ = ["complex_mtext_renderer"]


def corner_vertices(
    left: float,
    bottom: float,
    right: float,
    top: float,
    m: Matrix44 = None,
) -> Iterable[Vec3]:
    corners = [  # closed polygon: fist vertex  == last vertex
        (left, top),
        (right, top),
        (right, bottom),
        (left, bottom),
        (left, top),
    ]
    if m is None:
        return Vec3.generate(corners)
    else:
        return m.transform_vertices(corners)


class FrameRenderer(tl.ContentRenderer):
    def __init__(self, properties: Properties, backend: BackendInterface):
        self.properties = properties
        self.backend = backend

    def render(
        self,
        left: float,
        bottom: float,
        right: float,
        top: float,
        m: Matrix44 = None,
    ) -> None:
        self._render_outline(list(corner_vertices(left, bottom, right, top, m)))

    def _render_outline(self, vertices: List[Vec3]) -> None:
        backend = self.backend
        properties = self.properties
        prev = vertices.pop(0)
        for vertex in vertices:
            backend.draw_line(prev, vertex, properties)
            prev = vertex

    def line(
        self, x1: float, y1: float, x2: float, y2: float, m: Matrix44 = None
    ) -> None:
        points = [(x1, y1), (x2, y2)]
        if m is not None:
            p1, p2 = m.transform_vertices(points)
        else:
            p1, p2 = Vec3.generate(points)
        self.backend.draw_line(p1, p2, self.properties)


class ColumnBackgroundRenderer(FrameRenderer):
    def __init__(
        self,
        properties: Properties,
        backend: BackendInterface,
        bg_properties: Properties = None,
        offset: float = 0,
        text_frame: bool = False,
    ):
        super().__init__(properties, backend)
        self.bg_properties = bg_properties
        self.offset = offset  # background border offset
        self.has_text_frame = text_frame

    def render(
        self,
        left: float,
        bottom: float,
        right: float,
        top: float,
        m: Matrix44 = None,
    ) -> None:
        # Important: this is not a clipping box, it is possible to
        # render anything outside of the given borders!
        offset = self.offset
        vertices = list(
            corner_vertices(
                left - offset, bottom - offset, right + offset, top + offset, m
            )
        )
        if self.bg_properties is not None:
            self.backend.draw_filled_polygon(vertices, self.bg_properties)
        if self.has_text_frame:
            self._render_outline(vertices)


class TextRenderer(FrameRenderer):
    """Text content renderer."""

    def __init__(
        self,
        text: str,
        cap_height: float,
        width_factor: float,
        oblique: float,  # angle in degrees
        properties: Properties,
        backend: BackendInterface,
    ):
        super().__init__(properties, backend)
        self.text = text
        self.cap_height = cap_height
        self.width_factor = width_factor
        self.oblique = oblique  # angle in degrees

    def render(
        self,
        left: float,
        bottom: float,
        right: float,
        top: float,
        m: Matrix44 = None,
    ):
        """Create/render the text content"""
        sx = 1.0
        tx = 0.0
        if not math.isclose(self.width_factor, 1.0, rel_tol=1e-6):
            sx = self.width_factor
        if abs(self.oblique) > 1e-3:  # degrees
            tx = math.tan(math.radians(self.oblique))
        # fmt: off
        t = Matrix44((
            sx, 0.0, 0.0, 0.0,
            tx, 1.0, 0.0, 0.0,
            0.0, 0.0, 1.0, 0.0,
            left, bottom, 0.0, 1.0
        ))
        # fmt: on
        if m is not None:
            t *= m
        self.backend.draw_text(self.text, t, self.properties, self.cap_height)


def complex_mtext_renderer(
    ctx: RenderContext, backend: BackendInterface, mtext: MText, properties: Properties
) -> None:
    cmr = ComplexMTextRenderer(ctx, backend, properties)
    align = tl.LayoutAlignment(mtext.dxf.attachment_point)
    layout_engine = cmr.layout_engine(mtext)
    layout_engine.place(align=align)
    layout_engine.render(mtext.ucs().matrix)


class ComplexMTextRenderer(AbstractMTextRenderer):
    def __init__(
        self,
        ctx: RenderContext,
        backend: BackendInterface,
        properties: Properties,
    ):
        super().__init__()
        self._render_ctx = ctx
        self._backend = backend
        self._properties = properties

    # Implementation of required AbstractMTextRenderer methods:

    def word(self, text: str, ctx: MTextContext) -> tl.ContentCell:
        return tl.Text(
            width=self.get_font(ctx).text_width(text),
            height=ctx.cap_height,
            valign=tl.CellAlignment(ctx.align),
            stroke=self.get_stroke(ctx),
            renderer=TextRenderer(
                text,
                ctx.cap_height,
                ctx.width_factor,
                ctx.oblique,
                self.new_text_properties(self._properties, ctx),
                self._backend,
            ))

    def fraction(
        self, data: Tuple[str, str, str], ctx: MTextContext
    ) -> tl.ContentCell:
        upr, lwr, type_ = data
        if type_:
            return tl.Fraction(
                top=self.word(upr, ctx),
                bottom=self.word(lwr, ctx),
                stacking=self.get_stacking(type_),
                # renders just the divider line:
                renderer=FrameRenderer(self._properties, self._backend),
            )
        else:
            return self.word(upr, ctx)

    def get_font_face(self, mtext: MText) -> fonts.FontFace:
        return self._properties.font  # type: ignore

    def make_bg_renderer(self, mtext: MText) -> tl.ContentRenderer:
        dxf = mtext.dxf
        bg_fill = dxf.get("bg_fill", 0)

        bg_aci = None
        bg_true_color = None
        bg_properties: Optional[Properties] = None
        has_text_frame = False
        offset = 0
        if bg_fill:
            # The fill scale is a multiple of the initial char height and
            # a scale of 1, fits exact the outer border
            # of the column -> offset = 0
            offset = dxf.char_height * (dxf.get("box_fill_scale", 1.5) - 1)
            if bg_fill & const.MTEXT_BG_COLOR:
                if dxf.hasattr("bg_fill_color"):
                    bg_aci = dxf.bg_fill_color

                if dxf.hasattr("bg_fill_true_color"):
                    bg_aci = None
                    bg_true_color = dxf.bg_fill_true_color

                if (bg_fill & 3) == 3:  # canvas color = bit 0 and 1 set
                    # can not detect canvas color from DXF document!
                    # do not draw any background:
                    bg_aci = None
                    bg_true_color = None

            if bg_fill & const.MTEXT_TEXT_FRAME:
                has_text_frame = True
            bg_properties = self.new_bg_properties(bg_aci, bg_true_color)

        return ColumnBackgroundRenderer(
            self._properties,
            self._backend,
            bg_properties,
            offset=offset,
            text_frame=has_text_frame,
        )

    # Implementation details of ComplexMTextRenderer:

    @property
    def backend(self) -> BackendInterface:
        return self._backend

    def resolve_aci_color(self, aci: int) -> Color:
        return self._render_ctx.resolve_aci_color(aci, self._properties.layer)

    def new_text_properties(
        self, properties: Properties, ctx: MTextContext
    ) -> Properties:
        new_properties = copy.copy(properties)
        if ctx.rgb is None:
            new_properties.color = self.resolve_aci_color(ctx.aci)
        else:
            new_properties.color = rgb_to_hex(ctx.rgb)
        new_properties.font = ctx.font_face
        return new_properties

    def new_bg_properties(
        self, aci: Optional[int], true_color: Optional[int]
    ) -> Properties:
        new_properties = copy.copy(self._properties)
        new_properties.color = (  # canvas background color
            self._render_ctx.current_layout_properties.background_color
        )
        if true_color is None:
            if aci is not None:
                new_properties.color = self.resolve_aci_color(aci)
            # else canvas background color
        else:
            new_properties.color = rgb_to_hex(colors.int2rgb(true_color))
        return new_properties



# Purpose: menger sponge addon for ezdxf
# Copyright (c) 2016-2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, List, Tuple, Iterator
from ezdxf.math import Vec3
from ezdxf.render.mesh import MeshVertexMerger, MeshTransformer

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex, GenericLayoutType, Matrix44, UCS

# fmt: off
all_cubes_size_3_template = [
    (0, 0, 0), (1, 0, 0), (2, 0, 0), (0, 1, 0), (1, 1, 0), (2, 1, 0), (0, 2, 0), (1, 2, 0), (2, 2, 0),
    (0, 0, 1), (1, 0, 1), (2, 0, 1), (0, 1, 1), (1, 1, 1), (2, 1, 1), (0, 2, 1), (1, 2, 1), (2, 2, 1),
    (0, 0, 2), (1, 0, 2), (2, 0, 2), (0, 1, 2), (1, 1, 2), (2, 1, 2), (0, 2, 2), (1, 2, 2), (2, 2, 2),
]

original_menger_cubes = [
    (0, 0, 0), (1, 0, 0), (2, 0, 0), (0, 1, 0), (2, 1, 0), (0, 2, 0), (1, 2, 0), (2, 2, 0),
    (0, 0, 1), (2, 0, 1), (0, 2, 1), (2, 2, 1),
    (0, 0, 2), (1, 0, 2), (2, 0, 2), (0, 1, 2), (2, 1, 2), (0, 2, 2), (1, 2, 2), (2, 2, 2),
]

menger_v1 = [
    (0, 0, 0), (2, 0, 0), (1, 1, 0), (0, 2, 0), (2, 2, 0),
    (1, 0, 1), (0, 1, 1), (2, 1, 1), (1, 2, 1),
    (0, 0, 2), (2, 0, 2), (1, 1, 2), (0, 2, 2), (2, 2, 2),
]

menger_v2 = [
    (1, 0, 0), (0, 1, 0), (2, 1, 0), (1, 2, 0),
    (0, 0, 1), (2, 0, 1), (1, 1, 1), (0, 2, 1), (2, 2, 1),
    (1, 0, 2), (0, 1, 2), (2, 1, 2), (1, 2, 2),
]

jerusalem_cube = [
    (0, 0, 0), (1, 0, 0), (2, 0, 0), (3, 0, 0), (4, 0, 0), (0, 1, 0), (1, 1, 0), (3, 1, 0), (4, 1, 0), (0, 2, 0),
    (4, 2, 0), (0, 3, 0), (1, 3, 0), (3, 3, 0), (4, 3, 0), (0, 4, 0), (1, 4, 0), (2, 4, 0), (3, 4, 0), (4, 4, 0),
    (0, 0, 1), (1, 0, 1), (3, 0, 1), (4, 0, 1), (0, 1, 1), (1, 1, 1), (3, 1, 1), (4, 1, 1), (0, 3, 1), (1, 3, 1),
    (3, 3, 1), (4, 3, 1), (0, 4, 1), (1, 4, 1), (3, 4, 1), (4, 4, 1), (0, 0, 2), (4, 0, 2), (0, 4, 2), (4, 4, 2),
    (0, 0, 3), (1, 0, 3), (3, 0, 3), (4, 0, 3), (0, 1, 3), (1, 1, 3), (3, 1, 3), (4, 1, 3), (0, 3, 3), (1, 3, 3),
    (3, 3, 3), (4, 3, 3), (0, 4, 3), (1, 4, 3), (3, 4, 3), (4, 4, 3), (0, 0, 4), (1, 0, 4), (2, 0, 4), (3, 0, 4),
    (4, 0, 4), (0, 1, 4), (1, 1, 4), (3, 1, 4), (4, 1, 4), (0, 2, 4), (4, 2, 4), (0, 3, 4), (1, 3, 4), (3, 3, 4),
    (4, 3, 4), (0, 4, 4), (1, 4, 4), (2, 4, 4), (3, 4, 4), (4, 4, 4),
]

building_schemas = [
    original_menger_cubes,
    menger_v1,
    menger_v2,
    jerusalem_cube,
]

# subdivide level in order of building_schemas
cube_sizes = [3., 3., 3., 5.]

# 8 corner vertices
_cube_vertices = [
    (0, 0, 0),
    (1, 0, 0),
    (1, 1, 0),
    (0, 1, 0),
    (0, 0, 1),
    (1, 0, 1),
    (1, 1, 1),
    (0, 1, 1),
]

# 6 cube faces
cube_faces = [
    [0, 3, 2, 1],
    [4, 5, 6, 7],
    [0, 1, 5, 4],
    [1, 2, 6, 5],
    [3, 7, 6, 2],
    [0, 4, 7, 3],
]

# fmt: on


class MengerSponge:
    """

    Args:
        location: location of lower left corner as (x, y, z) tuple
        length: side length
        level: subdivide level
        kind: type of menger sponge

    === ===========================
    0   Original Menger Sponge
    1   Variant XOX
    2   Variant OXO
    3   Jerusalem Cube
    === ===========================

    """

    def __init__(
        self,
        location: "Vertex" = (0.0, 0.0, 0.0),
        length: float = 1.0,
        level: int = 1,
        kind: int = 0,
    ):
        self.cube_definitions = _menger_sponge(
            location=location, length=length, level=level, kind=kind
        )

    def vertices(self) -> Iterator["Vertex"]:
        """Yields the cube vertices as list of (x, y, z) tuples."""
        for location, length in self.cube_definitions:
            x, y, z = location
            yield [
                Vec3(x + xf * length, y + yf * length, z + zf * length)
                for xf, yf, zf in _cube_vertices
            ]

    __iter__ = vertices

    @staticmethod
    def faces() -> List[List[int]]:
        """Returns list of cube faces. All cube vertices have the same order, so
        one faces list fits them all.

        """
        return cube_faces

    def render(
        self,
        layout: "GenericLayoutType",
        merge: bool = False,
        dxfattribs: dict = None,
        matrix: "Matrix44" = None,
        ucs: "UCS" = None,
    ) -> None:
        """Renders the menger sponge into layout, set `merge` to ``True`` for
        rendering the whole menger sponge into one MESH entity, set `merge` to
        ``False`` for rendering the individual cubes of the menger sponge as
        MESH entities.

        Args:
            layout: DXF target layout
            merge: ``True`` for one MESH entity, ``False`` for individual MESH
                entities per cube
            dxfattribs: DXF attributes for the MESH entities
            matrix: apply transformation matrix at rendering
            ucs: apply UCS transformation at rendering

        """
        if merge:
            mesh = self.mesh()
            mesh.render_mesh(
                layout, dxfattribs=dxfattribs, matrix=matrix, ucs=ucs
            )
        else:
            for cube in self.cubes():
                cube.render_mesh(layout, dxfattribs, matrix=matrix, ucs=ucs)

    def cubes(self) -> Iterator[MeshTransformer]:
        """Yields all cubes of the menger sponge as individual
        :class:`MeshTransformer` objects.
        """
        faces = self.faces()
        for vertices in self:
            mesh = MeshVertexMerger()
            mesh.add_mesh(vertices=vertices, faces=faces)  # type: ignore
            yield MeshTransformer.from_builder(mesh)

    def mesh(self) -> MeshTransformer:
        """Returns geometry as one :class:`MeshTransformer` object."""
        faces = self.faces()
        mesh = MeshVertexMerger()
        for vertices in self:
            mesh.add_mesh(vertices=vertices, faces=faces)  # type: ignore
        return MeshTransformer.from_builder(mesh)


def _subdivide(
    location: "Vertex" = (0.0, 0.0, 0.0), length: float = 1.0, kind: int = 0
) -> List[Tuple["Vertex", float]]:
    """Divides a cube in sub-cubes and keeps only cubes determined by the
    building schema.

    All sides are parallel to x-, y- and z-axis, location is a (x, y, z) tuple
    and represents the coordinates of the lower left corner (nearest to the axis
    origin) of the cube, length is the side-length of the cube

    Args:
        location: (x, y, z) tuple, coordinates of the lower left corner of the cube
        length: side length of the cube
        kind: int for 0: original menger sponge; 1: Variant XOX; 2: Variant OXO;
            3: Jerusalem Cube;

    Returns: list of sub-cubes (location, length)

    """

    init_x, init_y, init_z = location
    step_size = float(length) / cube_sizes[kind]
    remaining_cubes = building_schemas[kind]

    def sub_location(indices) -> Vec3:
        x, y, z = indices
        return Vec3(
            init_x + x * step_size,
            init_y + y * step_size,
            init_z + z * step_size,
        )

    return [(sub_location(indices), step_size) for indices in remaining_cubes]


def _menger_sponge(
    location: "Vertex" = (0.0, 0.0, 0.0),
    length: float = 1.0,
    level: int = 1,
    kind: int = 0,
) -> List[Tuple[Vec3, float]]:
    """Builds a menger sponge for given level.

    Args:
        location: (x, y, z) tuple, coordinates of the lower left corner of the cube
        length: side length of the cube
        level: level of menger sponge, has to be 1 or bigger
        kind: int for 0: original menger sponge; 1: Variant XOX; 2: Variant OXO;
            3: Jerusalem Cube;

    Returns: list of sub-cubes (location, length)

    """
    kind = int(kind)
    if kind not in (0, 1, 2, 3):
        raise ValueError("kind has to be 0, 1, 2 or 3.")
    level = int(level)
    if level < 1:
        raise ValueError("level has to be 1 or bigger.")
    cubes = _subdivide(location, length, kind=kind)
    for _ in range(level - 1):
        next_level_cubes = []
        for location, length in cubes:
            next_level_cubes.extend(_subdivide(location, length, kind=kind))
        cubes = next_level_cubes
    return cubes
